#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from typing import Any, List, Dict
from functools import wraps
import logging

import numpy as np
import tensorcircuit as tc

try:
    import jax

    JAXIMPORTERROR = None
except ImportError as e:
    jax = None
    JAXIMPORTERROR = e

try:
    import cupy as cp
except ImportError:
    cp = None

Tensor = Any

logger = logging.getLogger(__name__)


ALL_JIT_LIBS: List[Dict] = []


def jit(f, static_argnums=None):
    backend_jit_lib = {}

    @wraps(f)
    def _wrap_jit(*args, **kwargs):
        if tc.backend.name not in backend_jit_lib:
            backend_jit_lib[tc.backend.name] = tc.backend.jit(f, static_argnums=static_argnums)
        return backend_jit_lib[tc.backend.name](*args, **kwargs)

    ALL_JIT_LIBS.append(backend_jit_lib)
    return _wrap_jit


def value_and_grad(f, argnums=0, has_aux=False):
    backend_vg_lib = {}

    @wraps(f)
    def _wrap_value_and_grad(*args, **kwargs):
        if tc.backend.name not in backend_vg_lib:
            try:
                vg_func = tc.backend.value_and_grad(f, argnums=argnums, has_aux=has_aux)
            except NotImplementedError:

                def vg_func(*args, **kwargs):
                    msg = (
                        f"The engine requires auto-differentiation, "
                        f"which is not available for backend {tc.backend.name}"
                    )
                    raise NotImplementedError(msg)

            backend_vg_lib[tc.backend.name] = vg_func
        return backend_vg_lib[tc.backend.name](*args, **kwargs)

    ALL_JIT_LIBS.append(backend_vg_lib)
    return _wrap_value_and_grad


set_backend = tc.set_backend


set_dtype = tc.set_dtype


def tensor_set_elem(tensor, idx, elem):
    if tc.backend.name == "jax":
        return tensor.at[idx].set(elem)
    else:
        assert tc.backend.name == "numpy" or tc.backend.name == "cupy"
        tensor[idx] = elem
        return tensor


def fori_loop(lower, upper, body_fun, init_val):
    if tc.backend.name == "jax":
        if jax is None:
            raise JAXIMPORTERROR
        return jax.lax.fori_loop(lower, upper, body_fun, init_val)

    assert tc.backend.name == "numpy" or tc.backend.name == "cupy"
    val = init_val
    for i in range(lower, upper):
        val = body_fun(i, val)
    return val


def scan(f, init, xs, length=None, reverse=False, unroll=1):
    if tc.backend.name == "jax":
        if jax is None:
            raise JAXIMPORTERROR
        return jax.lax.scan(f, init, xs, length, reverse, unroll)

    assert tc.backend.name == "numpy" or tc.backend.name == "cupy"
    carry = init
    ys = []
    xs = zip(*xs)
    if reverse:
        xs = reversed(list(xs))
    for x in xs:
        carry, y = f(carry, x)
        ys.append(y)
    return carry, np.stack(list(reversed(ys)))


def get_xp(_backend):
    if _backend.name == "cupy":
        return cp
    else:
        return np


def get_uint_type():
    if tc.rdtypestr == "float64":
        return np.uint64
    else:
        assert tc.rdtypestr == "float32"
        return np.uint32



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from functools import partial
from time import time
import logging
from typing import Any, Tuple, Callable, List, Union

import numpy as np
from scipy.optimize import minimize
from scipy.special import comb
import pandas as pd
from openfermion import jordan_wigner, FermionOperator, QubitOperator
from pyscf.gto.mole import Mole
from pyscf.scf import RHF
from pyscf.cc.addons import spatial2spin
from pyscf.mcscf import CASCI
from pyscf import fci
import tensorcircuit as tc

from tencirchem.constants import DISCARD_EPS
from tencirchem.molecule import _Molecule
from tencirchem.utils.misc import reverse_qop_idx, scipy_opt_wrap, rdm_mo2ao, canonical_mo_coeff
from tencirchem.utils.circuit import get_circuit_dataframe
from tencirchem.static.engine_ucc import (
    get_civector,
    get_statevector,
    get_energy,
    get_energy_and_grad,
    apply_excitation,
    translate_init_state,
)
from tencirchem.static.hamiltonian import (
    get_integral_from_hf,
    get_h_from_integral,
    get_hop_from_integral,
    get_hop_hcb_from_integral,
)
from tencirchem.static.ci_utils import get_ci_strings, get_ex_bitstring, get_addr, get_init_civector
from tencirchem.static.evolve_tensornetwork import get_circuit


logger = logging.getLogger(__name__)


Tensor = Any


class UCC:
    """
    Base class for :class:`UCCSD`.
    """

    @classmethod
    def from_integral(
        cls, int1e: np.ndarray, int2e: np.ndarray, n_elec: int, e_core: float = 0, ovlp: np.ndarray = None, **kwargs
    ):
        """
        Construct UCC classes from electron integrals.

        Parameters
        ----------
        int1e: np.ndarray
            One-body integral in spatial orbital.
        int2e: np.ndarray
            Two-body integral, in spatial orbital, chemists' notation, and without considering symmetry.
        n_elec: int
            The number of electrons
        e_core: float, optional
            The nuclear energy or core energy if active space approximation is involved.
            Defaults to 0.
        ovlp: np.ndarray
            The overlap integral. Defaults to ``None`` and identity matrix is used.
        kwargs:
            Other arguments to be passed to the :func:`__init__` function such as ``engine``.

        Returns
        -------
        ucc: :class:`UCC`
             A UCC instance
        """
        if isinstance(n_elec, tuple):
            if len(n_elec) != 2 or n_elec[0] != n_elec[1]:
                raise ValueError(f"Incompatible n_elec: {n_elec}")
            n_elec = n_elec[0] + n_elec[1]
        m = _Molecule(int1e, int2e, n_elec, e_core, ovlp)
        return cls(m, **kwargs)

    @classmethod
    def as_pyscf_solver(cls, config_function: Callable = None, **kwargs):
        """
        Converts the ``UCC`` class to a PySCF FCI solver.

        Parameters
        ----------
        config_function: callable
            A function to configure the ``UCC`` instance.
            Accepts the ``UCC`` instance and modifies it inplace before :func:`kernel` is called.
        kwargs
            Other arguments to be passed to the :func:`__init__` function such as ``engine``.

        Returns
        -------
        FCISolver

        Examples
        --------
        >>> from pyscf.mcscf import CASSCF
        >>> from tencirchem import UCCSD
        >>> from tencirchem.molecule import h8
        >>> # normal PySCF workflow
        >>> hf = h8.HF()
        >>> round(hf.kernel(), 8)
        -4.14961853
        >>> casscf = CASSCF(hf, 2, 2)
        >>> # set the FCI solver for CASSCF to be UCCSD
        >>> casscf.fcisolver = UCCSD.as_pyscf_solver()
        >>> round(casscf.kernel()[0], 8)
        -4.16647335
        """

        class FakeFCISolver:
            def __init__(self):
                self.instance: UCC = None
                self.config_function = config_function
                self.instance_kwargs = kwargs
                for arg in ["run_ccsd", "run_fci"]:
                    # keep MP2 for initial guess
                    self.instance_kwargs[arg] = False

            def kernel(self, h1, h2, norb, nelec, ci0=None, ecore=0, **kwargs):
                self.instance = cls.from_integral(h1, h2, nelec, **self.instance_kwargs)
                if self.config_function is not None:
                    self.config_function(self.instance)
                e = self.instance.kernel()
                return e + ecore, self.instance.params

            def make_rdm1(self, params, norb, nelec):
                civector = self.instance.civector(params)
                return self.instance.make_rdm1(civector)

            def make_rdm12(self, params, norb, nelec):
                civector = self.instance.civector(params)
                rdm1 = self.instance.make_rdm1(civector)
                rdm2 = self.instance.make_rdm2(civector)
                return rdm1, rdm2

            def spin_square(self, params, norb, nelec):
                return 0, 1

        return FakeFCISolver()

    def __init__(
        self,
        mol: Mole,
        init_method="mp2",
        active_space=None,
        mo_coeff=None,
        hcb=False,
        engine=None,
        run_hf=True,
        run_mp2=True,
        run_ccsd=True,
        run_fci=True,
    ):
        r"""
        Initialize the class with molecular input.

        Parameters
        ----------
        mol: Mole
            The molecule as PySCF ``Mole`` object.
        init_method: str, optional
            How to determine the initial amplitude guess. Accepts ``"mp2"`` (default), ``"ccsd"``
            and ``"zeros"``.
        active_space: Tuple[int, int], optional
            Active space approximation. The first integer is the number of electrons and the second integer is
            the number or spatial-orbitals. Defaults to None.
        mo_coeff: np.ndarray, optional
            Molecule coefficients. If provided then RHF is skipped.
            Can be used in combination with the ``init_state`` attribute.
            Defaults to None which means RHF orbitals are used.
        hcb: bool, optional
            Whether force electrons to pair as hard-core boson (HCB). Default to False.
        engine: str, optional
            The engine to run the calculation. See :ref:`advanced:Engines` for details.
        run_hf: bool, optional
            Whether run HF for molecule orbitals. Defaults to ``True``.
        run_mp2: bool, optional
            Whether run MP2 for initial guess and energy reference. Defaults to ``True``.
        run_ccsd: bool, optional
            Whether run CCSD for initial guess and energy reference. Defaults to ``True``.
        run_fci: bool, optional
            Whether run FCI  for energy reference. Defaults to ``True``.

        See Also
        --------
        tencirchem.UCCSD
        tencirchem.KUPCCGSD
        tencirchem.PUCCD
        """
        # process mol
        if isinstance(mol, _Molecule):
            self.mol = mol
        else:
            # to set verbose = 0
            self.mol = mol.copy()
            self.mol.symmetry = True
            self.mol.build()
        if active_space is None:
            active_space = (mol.nelectron, int(mol.nao))

        self.hcb = hcb
        self.n_qubits = 2 * active_space[1]
        if hcb:
            self.n_qubits //= 2

        # process activate space
        self.active_space = active_space
        self.n_elec = active_space[0]
        self.active = active_space[1]
        self.inactive_occ = mol.nelectron // 2 - active_space[0] // 2
        self.inactive_vir = mol.nao - active_space[1] - self.inactive_occ
        frozen_idx = list(range(self.inactive_occ)) + list(range(mol.nao - self.inactive_vir, mol.nao))

        # process backend
        self._check_engine(engine)

        if engine is None:
            # no need to be too precise
            if self.n_qubits <= 16:
                engine = "civector"
            else:
                engine = "civector-large"
        self.engine = engine

        # classical quantum chemistry
        # hf
        self.mol.verbose = 0
        self.hf = RHF(self.mol)
        # avoid serialization warnings for `_Molecule`
        self.hf.chkfile = None
        if run_hf:
            # run this even when ``mo_coeff is not None`` because MP2 and CCSD
            # reference energy might be desired
            self.e_hf = self.hf.kernel()
            self.hf.mo_coeff = canonical_mo_coeff(self.hf.mo_coeff)
        else:
            self.e_hf = None
            # otherwise, can't run casci.get_h2eff() based on HF
            self.hf._eri = mol.intor("int2e", aosym="s8")
            if mo_coeff is None:
                raise ValueError("Must provide MO coefficient if HF is skipped")

        # mp2
        if run_mp2:
            mp2 = self.hf.MP2()
            if frozen_idx:
                mp2.frozen = frozen_idx
            e_corr_mp2, mp2_t2 = mp2.kernel()
            self.e_mp2 = self.e_hf + e_corr_mp2
        else:
            self.e_mp2 = None
            mp2_t2 = None
            if init_method is not None and init_method.lower() == "mp2":
                raise ValueError("Must run MP2 to use MP2 as the initial guess method")

        # ccsd
        if run_ccsd:
            ccsd = self.hf.CCSD()
            if frozen_idx:
                ccsd.frozen = frozen_idx
            e_corr_ccsd, ccsd_t1, ccsd_t2 = ccsd.kernel()
            self.e_ccsd = self.e_hf + e_corr_ccsd
        else:
            self.e_ccsd = None
            ccsd_t1 = ccsd_t2 = None
            if init_method is not None and init_method.lower() == "ccsd":
                raise ValueError("Must run CCSD to use CCSD as the initial guess method")

        # MP2 and CCSD rely on canonical HF orbitals but FCI doesn't
        # so set custom mo_coeff after MP2 and CCSD and before FCI
        if mo_coeff is not None:
            # use user defined coefficient
            self.hf.mo_coeff = canonical_mo_coeff(mo_coeff)

        # fci
        if run_fci:
            fci = CASCI(self.hf, self.active_space[1], self.active_space[0])
            fci.max_memory = 32000
            res = fci.kernel()
            self.e_fci = res[0]
            self.civector_fci = res[2].ravel()
        else:
            self.e_fci = None
            self.civector_fci = None

        self.e_nuc = mol.energy_nuc()

        # initial guess
        self.t1 = self.t2 = None
        if init_method is None or init_method in ["zeros", "zero"]:
            pass
        elif init_method.lower() == "ccsd":
            self.t1, self.t2 = ccsd_t1, ccsd_t2
        elif init_method.lower() == "mp2":
            self.t2 = mp2_t2
        else:
            raise ValueError(f"Unknown initialization method: {init_method}")

        # Hamiltonian related
        self.hamiltonian_lib = {}
        self.int1e = self.int2e = None
        # e_core includes nuclear repulsion energy
        self.hamiltonian, self.e_core, _ = self._get_hamiltonian_and_core(self.engine)

        # circuit related
        self._init_state = None
        self.ex_ops = None
        self._param_ids = None
        self.init_guess = None

        # optimization related
        self.scipy_minimize_options = None
        # optimization result
        self.opt_res = None
        # for manually set
        self._params = None

    def kernel(self) -> float:
        """
        The kernel to perform the VQE algorithm.
        The L-BFGS-B method in SciPy is used for optimization
        and configuration is possible by setting the ``self.scipy_minimize_options`` attribute.

        Returns
        -------
        e: float
            The optimized energy
        """
        assert len(self.param_ids) == len(self.ex_ops)

        energy_and_grad, stating_time = self.get_opt_function(with_time=True)

        if self.init_guess is None:
            self.init_guess = np.zeros(self.n_params)

        # optimization options
        if self.scipy_minimize_options is None:
            # quite strict
            options = {"ftol": 1e1 * np.finfo(tc.rdtypestr).eps, "gtol": 1e2 * np.finfo(tc.rdtypestr).eps}
        else:
            options = self.scipy_minimize_options

        logger.info("Begin optimization")

        time1 = time()
        opt_res = minimize(energy_and_grad, x0=self.init_guess, jac=True, method="L-BFGS-B", options=options)
        time2 = time()

        if not opt_res.success:
            logger.warning("Optimization failed. See `.opt_res` for details.")

        opt_res["staging_time"] = stating_time
        opt_res["opt_time"] = time2 - time1
        opt_res["init_guess"] = self.init_guess
        opt_res["e"] = float(opt_res.fun)
        self.opt_res = opt_res
        # prepare for future modification
        self.params = opt_res.x.copy()
        return opt_res.e

    def get_opt_function(self, with_time: bool = False) -> Union[Callable, Tuple[Callable, float]]:
        """
        Returns the cost function in SciPy format for optimization.
        The gradient is included.
        Basically a wrapper to :func:`energy_and_grad`.

        Parameters
        ----------
        with_time: bool, optional
            Whether return staging time. Defaults to False.

        Returns
        -------
        opt_function: Callable
            The optimization cost function in SciPy format.
        time: float
            Staging time. Returned when ``with_time`` is set to ``True``.
        """
        energy_and_grad = scipy_opt_wrap(partial(self.energy_and_grad, engine=self.engine))

        time1 = time()
        if tc.backend.name == "jax":
            logger.info("JIT compiling the circuit")
            _ = energy_and_grad(np.zeros(self.n_params))
            logger.info("Circuit JIT compiled")
        time2 = time()
        if with_time:
            return energy_and_grad, time2 - time1
        return energy_and_grad

    def _check_params_argument(self, params, strict=True):
        if params is None:
            if self.params is not None:
                params = self.params
            else:
                if strict:
                    raise ValueError("Run the `.kernel` method to determine the parameters first")
                else:
                    if self.init_guess is not None:
                        params = self.init_guess
                    else:
                        params = np.zeros(self.n_params)

        if len(params) != self.n_params:
            raise ValueError(f"Incompatible parameter shape. {self.n_params} is desired. Got {len(params)}")
        return tc.backend.convert_to_tensor(params).astype(tc.rdtypestr)

    def _check_engine(self, engine):
        supported_engine = [None, "tensornetwork", "statevector", "civector", "civector-large", "pyscf"]
        if not engine in supported_engine:
            raise ValueError(f"Engine '{engine}' not supported")

    def _sanity_check(self):
        if self.ex_ops is None or self.param_ids is None:
            raise ValueError("`ex_ops` or `param_ids` not defined")
        if self.param_ids is not None and (len(self.ex_ops) != len(self.param_ids)):
            raise ValueError(
                f"Excitation operator size {len(self.ex_ops)} and parameter size {len(self.param_ids)} do not match"
            )

    def civector(self, params: Tensor = None, engine: str = None) -> Tensor:
        """
        Evaluate the configuration interaction (CI) vector.

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.
        engine: str, optional
            The engine to use. Defaults to ``None``, which uses ``self.engine``.

        Returns
        -------
        civector: Tensor
            Corresponding CI vector

        See Also
        --------
        statevector: Evaluate the circuit state vector.
        energy: Evaluate the total energy.
        energy_and_grad: Evaluate the total energy and parameter gradients.

        Examples
        --------
        >>> from tencirchem import UCCSD
        >>> from tencirchem.molecule import h2
        >>> uccsd = UCCSD(h2)
        >>> uccsd.civector([0, 0])  # HF state
        array([1., 0., 0., 0.])
        """
        self._sanity_check()
        params = self._check_params_argument(params)
        self._check_engine(engine)
        if engine is None:
            engine = self.engine
        civector = get_civector(
            params, self.n_qubits, self.n_elec, self.ex_ops, self.param_ids, self.hcb, self.init_state, engine
        )
        return civector

    def get_ci_strings(self, strs2addr: bool = False) -> np.ndarray:
        """
        Get the CI bitstrings for all configurations in the CI vector.

        Parameters
        ----------
        strs2addr: bool, optional.
            Whether return the reversed mapping for one spin sector. Defaults to ``False``.

        Returns
        -------
        cistrings: np.ndarray
            The CI bitstrings.
        string_addr: np.ndarray
            The address of the string in one spin sector. Returned when ``strs2addr`` is set to ``True``.

        Examples
        --------
        >>> from tencirchem import UCCSD, PUCCD
        >>> from tencirchem.molecule import h2
        >>> uccsd = UCCSD(h2)
        >>> uccsd.get_ci_strings()
        array([ 5,  6,  9, 10], dtype=uint64)
        >>> [f"{bin(i)[2:]:0>4}" for i in uccsd.get_ci_strings()]
        ['0101', '0110', '1001', '1010']
        >>> uccsd.get_ci_strings(True)[1]  # only one spin sector
        array([0, 0, 1, 0], dtype=uint64)
        """
        return get_ci_strings(self.n_qubits, self.n_elec, self.hcb, strs2addr=strs2addr)

    def get_addr(self, bitstring: str) -> int:
        """
        Get the address (index) of a CI bitstring in the CI vector.

        Parameters
        ----------
        bitstring: str
            The bitstring such as ``"0101"``.

        Returns
        -------
        address: int
            The bitstring address.

        Examples
        --------
        >>> from tencirchem import UCCSD, PUCCD
        >>> from tencirchem.molecule import h2
        >>> uccsd = UCCSD(h2)
        >>> uccsd.get_addr("0101")  # the HF state
        0
        >>> uccsd.get_addr("1010")
        3
        >>> puccd = PUCCD(h2)
        >>> puccd.get_addr("01")  # the HF state
        0
        >>> puccd.get_addr("10")
        1
        """
        _, strs2addr = self.get_ci_strings(strs2addr=True)
        return int(get_addr(int(bitstring, base=2), self.n_qubits, self.n_elec, strs2addr, self.hcb))

    def statevector(self, params: Tensor = None, engine: str = None) -> Tensor:
        """
        Evaluate the circuit state vector.

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.
        engine: str, optional
            The engine to use. Defaults to ``None``, which uses ``self.engine``.

        Returns
        -------
        statevector: Tensor
            Corresponding state vector

        See Also
        --------
        civector: Evaluate the configuration interaction (CI) vector.
        energy: Evaluate the total energy.
        energy_and_grad: Evaluate the total energy and parameter gradients.

        Examples
        --------
        >>> from tencirchem import UCCSD
        >>> from tencirchem.molecule import h2
        >>> uccsd = UCCSD(h2)
        >>> uccsd.statevector([0, 0])  # HF state
        array([0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.])
        """
        self._sanity_check()
        params = self._check_params_argument(params)
        self._check_engine(engine)
        if engine is None:
            engine = self.engine
        statevector = get_statevector(
            params, self.n_qubits, self.n_elec, self.ex_ops, self.param_ids, self.hcb, self.init_state, engine
        )
        return statevector

    def _get_hamiltonian_and_core(self, engine):
        self._check_engine(engine)
        if engine is None:
            engine = self.engine
            hamiltonian = self.hamiltonian
            e_core = self.e_core
        else:
            if engine.startswith("civector") or engine == "pyscf":
                htype = "fcifunc"
            else:
                assert engine in ["tensornetwork", "statevector"]
                htype = "sparse"
            hamiltonian = self.hamiltonian_lib.get(htype)
            if hamiltonian is None:
                if self.int1e is None:
                    self.int1e, self.int2e, e_core = get_integral_from_hf(self.hf, self.active_space)
                else:
                    e_core = self.e_core
                hamiltonian = get_h_from_integral(self.int1e, self.int2e, self.n_elec, self.hcb, htype)
                self.hamiltonian_lib[htype] = hamiltonian
            else:
                e_core = self.e_core
        return hamiltonian, e_core, engine

    def energy(self, params: Tensor = None, engine: str = None) -> float:
        """
        Evaluate the total energy.

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.
        engine: str, optional
            The engine to use. Defaults to ``None``, which uses ``self.engine``.

        Returns
        -------
        energy: float
            Total energy

        See Also
        --------
        civector: Get the configuration interaction (CI) vector.
        statevector: Evaluate the circuit state vector.
        energy_and_grad: Evaluate the total energy and parameter gradients.

        Examples
        --------
        >>> from tencirchem import UCCSD
        >>> from tencirchem.molecule import h2
        >>> uccsd = UCCSD(h2)
        >>> round(uccsd.energy([0, 0]), 8)  # HF state
        -1.11670614
        """
        self._sanity_check()
        params = self._check_params_argument(params)
        if params is self.params and self.opt_res is not None:
            return self.opt_res.e
        hamiltonian, _, engine = self._get_hamiltonian_and_core(engine)
        e = get_energy(
            params,
            hamiltonian,
            self.n_qubits,
            self.n_elec,
            self.ex_ops,
            self.param_ids,
            self.hcb,
            self.init_state,
            engine,
        )
        return float(e) + self.e_core

    def energy_and_grad(self, params: Tensor = None, engine: str = None) -> Tuple[float, Tensor]:
        """
        Evaluate the total energy and parameter gradients.

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.
        engine: str, optional
            The engine to use. Defaults to ``None``, which uses ``self.engine``.

        Returns
        -------
        energy: float
            Total energy
        grad: Tensor
            The parameter gradients

        See Also
        --------
        civector: Get the configuration interaction (CI) vector.
        statevector: Evaluate the circuit state vector.
        energy: Evaluate the total energy.

        Examples
        --------
        >>> from tencirchem import UCCSD
        >>> from tencirchem.molecule import h2
        >>> uccsd = UCCSD(h2)
        >>> e, g = uccsd.energy_and_grad([0, 0])
        >>> round(e, 8)
        -1.11670614
        >>> g  # doctest:+ELLIPSIS
        array([..., ...])
        """
        self._sanity_check()
        params = self._check_params_argument(params)
        hamiltonian, _, engine = self._get_hamiltonian_and_core(engine)
        e, g = get_energy_and_grad(
            params,
            hamiltonian,
            self.n_qubits,
            self.n_elec,
            self.ex_ops,
            self.param_ids,
            self.hcb,
            self.init_state,
            engine,
        )
        return float(e + self.e_core), tc.backend.numpy(g)

    def apply_excitation(self, state: Tensor, ex_op: Tuple, engine: str = None) -> Tensor:
        """
        Apply a given excitation operator to a given state.

        Parameters
        ----------
        state: Tensor
            The input state in statevector or CI vector form.
        ex_op: tuple of ints
            The excitation operator.
        engine: str, optional
            The engine to use. Defaults to ``None``, which uses ``self.engine``.
        Returns
        -------
        tensor: Tensor
            The resulting tensor.

        Examples
        --------
        >>> from tencirchem import UCC
        >>> from tencirchem.molecule import h2
        >>> ucc = UCC(h2)
        >>> ucc.apply_excitation([1, 0, 0, 0], (3, 1, 0, 2))
        array([0, 0, 0, 1])
        """
        self._check_engine(engine)
        if engine is None:
            engine = self.engine
        return apply_excitation(state, self.n_qubits, self.n_elec, ex_op, hcb=self.hcb, engine=engine)

    def _statevector_to_civector(self, statevector=None):
        if statevector is None:
            civector = self.civector()
        else:
            if len(statevector) == self.statevector_size:
                ci_strings = self.get_ci_strings()
                civector = statevector[ci_strings]
            else:
                if len(statevector) == self.civector_size:
                    civector = statevector
                else:
                    raise ValueError(f"Incompatible statevector size: {len(statevector)}")

        civector = tc.backend.numpy(tc.backend.convert_to_tensor(civector))
        return civector

    def make_rdm1(self, statevector: Tensor = None, basis: str = "AO") -> np.ndarray:
        r"""
        Evaluate the spin-traced one-body reduced density matrix (1RDM).

        .. math::

            \textrm{1RDM}[p,q] = \langle p_{\alpha}^\dagger q_{\alpha} \rangle
                + \langle p_{\beta}^\dagger q_{\beta} \rangle

        If active space approximation is employed, returns the full RDM of all orbitals.

        Parameters
        ----------
        statevector: Tensor, optional
            Custom system state. Could be CI vector or state vector.
            Defaults to None, which uses the optimized state by :func:`civector`.

        basis: str, optional
            One of ``"AO"`` or ``"MO"``. Defaults to ``"AO"``, which is for consistency with PySCF.

        Returns
        -------
        rdm1: np.ndarray
            The spin-traced one-body RDM.

        See Also
        --------
        make_rdm2: Evaluate the spin-traced two-body reduced density matrix (2RDM).
        """
        assert not self.hcb
        civector = self._statevector_to_civector(statevector).astype(np.float64)

        rdm1_cas = fci.direct_spin1.make_rdm1(civector, self.n_qubits // 2, self.n_elec)

        rdm1 = self.embed_rdm_cas(rdm1_cas)

        if basis == "MO":
            return rdm1
        else:
            return rdm_mo2ao(rdm1, self.hf.mo_coeff)

    def make_rdm2(self, statevector: Tensor = None, basis: str = "AO") -> np.ndarray:
        r"""
        Evaluate the spin-traced two-body reduced density matrix (1RDM).

        .. math::

            \begin{aligned}
                \textrm{2RDM}[p,q,r,s] & = \langle p_{\alpha}^\dagger r_{\alpha}^\dagger
                s_{\alpha}  q_{\alpha} \rangle
                   + \langle p_{\beta}^\dagger r_{\alpha}^\dagger s_{\alpha}  q_{\beta} \rangle \\
                   & \quad + \langle p_{\alpha}^\dagger r_{\beta}^\dagger s_{\beta}  q_{\alpha} \rangle
                   + \langle p_{\beta}^\dagger r_{\beta}^\dagger s_{\beta}  q_{\beta} \rangle
            \end{aligned}

        If active space approximation is employed, returns the full RDM of all orbitals.

        Parameters
        ----------
        statevector: Tensor, optional
            Custom system state. Could be CI vector or state vector.
            Defaults to None, which uses the optimized state by :func:`civector`.

        basis: str, optional
            One of ``"AO"`` or ``"MO"``. Defaults to ``"AO"``, which is for consistency with PySCF.

        Returns
        -------
        rdm2: np.ndarray
            The spin-traced two-body RDM.

        See Also
        --------
        make_rdm1: Evaluate the spin-traced one-body reduced density matrix (1RDM).

        Examples
        --------
        >>> import numpy as np
        >>> from tencirchem import UCC
        >>> from tencirchem.molecule import h2
        >>> ucc = UCC(h2)
        >>> state = [1, 0, 0, 0]  ## HF state
        >>> rdm1 = ucc.make_rdm1(state, basis="MO")
        >>> rdm2 = ucc.make_rdm2(state, basis="MO")
        >>> e_hf = ucc.int1e.ravel() @ rdm1.ravel() + 1/2 * ucc.int2e.ravel() @ rdm2.ravel()
        >>> np.testing.assert_allclose(e_hf + ucc.e_nuc, ucc.e_hf, atol=1e-10)
        """
        assert not self.hcb
        civector = self._statevector_to_civector(statevector).astype(np.float64)

        rdm2_cas = fci.direct_spin1.make_rdm12(civector.astype(np.float64), self.n_qubits // 2, self.n_elec)[1]

        rdm2 = self.embed_rdm_cas(rdm2_cas)

        if basis == "MO":
            return rdm2
        else:
            return rdm_mo2ao(rdm2, self.hf.mo_coeff)

    def embed_rdm_cas(self, rdm_cas):
        """
        Embed CAS RDM into RDM of the whole system
        """
        if self.inactive_occ == 0 and self.inactive_vir == 0:
            # active space approximation not employed
            return rdm_cas
        # slice of indices in rdm corresponding to cas
        slice_cas = slice(self.inactive_occ, self.inactive_occ + len(rdm_cas))
        if rdm_cas.ndim == 2:
            rdm1_cas = rdm_cas
            rdm1 = np.zeros((self.mol.nao, self.mol.nao))
            for i in range(self.inactive_occ):
                rdm1[i, i] = 2
            rdm1[slice_cas, slice_cas] = rdm1_cas
            return rdm1
        else:
            rdm2_cas = rdm_cas
            # active space approximation employed
            rdm1 = self.make_rdm1(basis="MO")
            rdm1_cas = rdm1[slice_cas, slice_cas]
            rdm2 = np.zeros((self.mol.nao, self.mol.nao, self.mol.nao, self.mol.nao))
            rdm2[slice_cas, slice_cas, slice_cas, slice_cas] = rdm2_cas
            for i in range(self.inactive_occ):
                for j in range(self.inactive_occ):
                    rdm2[i, i, j, j] += 4
                    rdm2[i, j, j, i] -= 2
                rdm2[i, i, slice_cas, slice_cas] = rdm2[slice_cas, slice_cas, i, i] = 2 * rdm1_cas
                rdm2[i, slice_cas, slice_cas, i] = rdm2[slice_cas, i, i, slice_cas] = -rdm1_cas
            return rdm2

    def get_ex_ops(self, t1: np.ndarray = None, t2: np.ndarray = None):
        """Virtual method to be implemented"""
        raise NotImplementedError

    def get_ex1_ops(self, t1: np.ndarray = None) -> Tuple[List[Tuple], List[int], List[float]]:
        """
        Get one-body excitation operators.

        Parameters
        ----------
        t1: np.ndarray, optional
            Initial one-body amplitudes based on e.g. CCSD

        Returns
        -------
        ex_op: List[Tuple]
            The excitation operators. Each operator is represented by a tuple of ints.
        param_ids: List[int]
            The mapping from excitations to parameters.
        init_guess: List[float]
            The initial guess for the parameters.

        See Also
        --------
        get_ex2_ops: Get two-body excitation operators.
        get_ex_ops: Get one-body and two-body excitation operators for UCCSD ansatz.
        """
        # single excitations
        no, nv = self.no, self.nv
        if t1 is None:
            t1 = np.zeros((no, nv))

        t1 = spatial2spin(t1)

        ex1_ops = []
        # unique parameters. -1 is a place holder
        ex1_param_ids = [-1]
        ex1_init_guess = []
        for i in range(no):
            for a in range(nv):
                # alpha to alpha
                ex_op_a = (2 * no + nv + a, no + nv + i)
                # beta to beta
                ex_op_b = (no + a, i)
                ex1_ops.extend([ex_op_a, ex_op_b])
                ex1_param_ids.extend([ex1_param_ids[-1] + 1] * 2)
                ex1_init_guess.append(t1[i, a])

        return ex1_ops, ex1_param_ids[1:], ex1_init_guess

    def get_ex2_ops(self, t2: np.ndarray = None) -> Tuple[List[Tuple], List[int], List[float]]:
        """
        Get two-body excitation operators.

        Parameters
        ----------
        t2: np.ndarray, optional
            Initial two-body amplitudes based on e.g. MP2

        Returns
        -------
        ex_op: List[Tuple]
            The excitation operators. Each operator is represented by a tuple of ints.
        param_ids: List[int]
            The mapping from excitations to parameters.
        init_guess: List[float]
            The initial guess for the parameters.

        See Also
        --------
        get_ex1_ops: Get one-body excitation operators.
        get_ex_ops: Get one-body and two-body excitation operators for UCCSD ansatz.
        """

        # t2 in oovv 1212 format
        no, nv = self.no, self.nv
        if t2 is None:
            t2 = np.zeros((no, no, nv, nv))

        t2 = spatial2spin(t2)

        def alpha_o(_i):
            return no + nv + _i

        def alpha_v(_i):
            return 2 * no + nv + _i

        def beta_o(_i):
            return _i

        def beta_v(_i):
            return no + _i

        # double excitations
        ex_ops = []
        ex2_param_ids = [-1]
        ex2_init_guess = []
        # 2 alphas or 2 betas
        for i in range(no):
            for j in range(i):
                for a in range(nv):
                    for b in range(a):
                        # i correspond to a and j correspond to b, as in PySCF convention
                        # otherwise the t2 amplitude has incorrect phase
                        # 2 alphas
                        ex_op_aa = (alpha_v(b), alpha_v(a), alpha_o(i), alpha_o(j))
                        # 2 betas
                        ex_op_bb = (beta_v(b), beta_v(a), beta_o(i), beta_o(j))
                        ex_ops.extend([ex_op_aa, ex_op_bb])
                        ex2_param_ids.extend([ex2_param_ids[-1] + 1] * 2)
                        ex2_init_guess.append(t2[2 * i, 2 * j, 2 * a, 2 * b])
        assert len(ex_ops) == 2 * (no * (no - 1) / 2) * (nv * (nv - 1) / 2)
        # 1 alpha + 1 beta
        for i in range(no):
            for j in range(i + 1):
                for a in range(nv):
                    for b in range(a + 1):
                        # i correspond to a and j correspond to b, as in PySCF convention
                        # otherwise the t2 amplitude has incorrect phase
                        if i == j and a == b:
                            # paired
                            ex_op_ab = (beta_v(a), alpha_v(a), alpha_o(i), beta_o(i))
                            ex_ops.append(ex_op_ab)
                            ex2_param_ids.append(ex2_param_ids[-1] + 1)
                            ex2_init_guess.append(t2[2 * i, 2 * i + 1, 2 * a, 2 * a + 1])
                            continue
                        # simple reflection
                        ex_op_ab1 = (beta_v(b), alpha_v(a), alpha_o(i), beta_o(j))
                        ex_op_ab2 = (alpha_v(b), beta_v(a), beta_o(i), alpha_o(j))
                        ex_ops.extend([ex_op_ab1, ex_op_ab2])
                        ex2_param_ids.extend([ex2_param_ids[-1] + 1] * 2)
                        ex2_init_guess.append(t2[2 * i, 2 * j + 1, 2 * a, 2 * b + 1])
                        if (i != j) and (a != b):
                            # exchange alpha and beta
                            ex_op_ab3 = (beta_v(a), alpha_v(b), alpha_o(i), beta_o(j))
                            ex_op_ab4 = (alpha_v(a), beta_v(b), beta_o(i), alpha_o(j))
                            ex_ops.extend([ex_op_ab3, ex_op_ab4])
                            ex2_param_ids.extend([ex2_param_ids[-1] + 1] * 2)
                            ex2_init_guess.append(t2[2 * i, 2 * j + 1, 2 * b, 2 * a + 1])

        return ex_ops, ex2_param_ids[1:], ex2_init_guess

    @property
    def e_ucc(self) -> float:
        """
        Returns UCC energy
        """
        return self.energy()

    def print_ansatz(self):
        df_dict = {
            "#qubits": [self.n_qubits],
            "#params": [self.n_params],
            "#excitations": [len(self.ex_ops)],
        }
        if self.init_state is None:
            df_dict["initial condition"] = "RHF"
        else:
            df_dict["initial condition"] = "custom"
        print(pd.DataFrame(df_dict).to_string(index=False))

    def get_circuit(
        self, params: Tensor = None, decompose_multicontrol: bool = False, trotter: bool = False
    ) -> tc.Circuit:
        """
        Get the circuit as TensorCircuit ``Circuit`` object

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter.
            If :func:`kernel` is not called before, the initial guess is used.
        decompose_multicontrol: bool, optional
            Whether decompose the Multicontrol gate in the circuit into CNOT gates.
            Defaults to False.
        trotter: bool, optional
            Whether Trotterize the UCC factor into Pauli strings.
            Defaults to False.

        Returns
        -------
        circuit: :class:`tc.Circuit`
            The quantum circuit.
        """
        if self.ex_ops is None:
            raise ValueError("Excitation operators not defined")
        params = self._check_params_argument(params, strict=False)
        return get_circuit(
            params,
            self.n_qubits,
            self.n_elec,
            self.ex_ops,
            self.param_ids,
            self.hcb,
            self.init_state,
            decompose_multicontrol=decompose_multicontrol,
            trotter=trotter,
        )

    def print_circuit(self):
        """
        Prints the circuit information. If you wish to print the circuit diagram,
        use :func:`get_circuit` and then call ``draw()`` such as ``print(ucc.get_circuit().draw())``.
        """
        c = self.get_circuit()
        df = get_circuit_dataframe(c)

        def format_flop(f):
            return f"{f:.3e}"

        formatters = {"flop": format_flop}
        print(df.to_string(index=False, formatters=formatters))

    def get_init_state_dataframe(self, coeff_epsilon: float = DISCARD_EPS) -> pd.DataFrame:
        """
        Returns initial state information dataframe.

        Parameters
        ----------
        coeff_epsilon: float, optional
            The threshold to screen out states with small coefficients.
            Defaults to 1e-12.

        Returns
        -------
        pd.DataFrame

        See Also
        --------
        init_state: The circuit initial state before applying the excitation operators.

        Examples
        --------
        >>> from tencirchem import UCC
        >>> from tencirchem.molecule import h2
        >>> ucc = UCC(h2)
        >>> ucc.init_state = [0.707, 0, 0, 0.707]
        >>> ucc.get_init_state_dataframe()   # doctest: +NORMALIZE_WHITESPACE
             configuration  coefficient
        0          0101        0.707
        1          1010        0.707
        """
        columns = ["configuration", "coefficient"]
        if self.init_state is None:
            init_state = get_init_civector(self.civector_size)
        else:
            init_state = self.init_state
        ci_strings = self.get_ci_strings()
        ci_coeffs = translate_init_state(init_state, self.n_qubits, ci_strings)
        data_list = []
        for ci_string, coeff in zip(ci_strings, ci_coeffs):
            if np.abs(coeff) < coeff_epsilon:
                continue
            ci_string = bin(ci_string)[2:]
            ci_string = "0" * (self.n_qubits - len(ci_string)) + ci_string
            data_list.append((ci_string, coeff))
        return pd.DataFrame(data_list, columns=columns)

    def print_init_state(self):
        print(self.get_init_state_dataframe().to_string())

    def get_excitation_dataframe(self) -> pd.DataFrame:
        columns = ["excitation", "configuration", "parameter", "initial guess"]
        if self.ex_ops is None:
            return pd.DataFrame(columns=columns)

        if self.params is None:
            # optimization not done
            params = [None] * len(self.init_guess)
        else:
            params = self.params

        if self.param_ids is None:
            # see self.n_params
            param_ids = range(len(self.ex_ops))
        else:
            param_ids = self.param_ids

        data_list = []

        for i, ex_op in zip(param_ids, self.ex_ops):
            bitstring = get_ex_bitstring(self.n_qubits, self.n_elec, ex_op, self.hcb)
            data_list.append((ex_op, bitstring, params[i], self.init_guess[i]))
        return pd.DataFrame(data_list, columns=columns)

    def print_excitations(self):
        print(self.get_excitation_dataframe().to_string())

    def get_energy_dataframe(self) -> pd.DataFrame:
        """
        Returns energy information dataframe
        """
        if self.params is None:
            series_dict = {"HF": self.e_hf, "MP2": self.e_mp2, "CCSD": self.e_ccsd, "FCI": self.e_fci}
        else:
            ucc_name = self.__class__.__name__
            series_dict = {
                "HF": self.e_hf,
                "MP2": self.e_mp2,
                "CCSD": self.e_ccsd,
                ucc_name: self.energy(),
                "FCI": self.e_fci,
            }
        df = pd.DataFrame()
        energy = pd.Series(series_dict)
        df["energy (Hartree)"] = energy
        if self.e_fci is not None:
            df["error (mH)"] = 1e3 * (energy - self.e_fci)
            df["correlation energy (%)"] = 100 * (energy - self.e_hf) / (self.e_fci - self.e_hf)
        return df

    def print_energy(self):
        df = self.get_energy_dataframe()

        def format_ce(f):
            return f"{f:.3f}"

        formatters = {"correlation energy (%)": format_ce}
        print(df.to_string(index=True, formatters=formatters))

    def print_summary(self, include_circuit: bool = False):
        """
        Print a summary of the class.

        Parameters
        ----------
        include_circuit: bool
            Whether include the circuit section.

        """
        print("################################ Ansatz ###############################")
        self.print_ansatz()
        if self.init_state is not None:
            print("############################ Initial Condition ########################")
            self.print_init_state()
        if include_circuit:
            print("############################### Circuit ###############################")
            self.print_circuit()
        print("############################### Energy ################################")
        self.print_energy()
        print("############################# Excitations #############################")
        self.print_excitations()
        print("######################### Optimization Result #########################")
        if self.opt_res is None:
            print("Optimization not run (.opt_res is None)")
        else:
            print(self.opt_res)

    @property
    def no(self) -> int:
        """The number of occupied orbitals."""
        return self.n_elec // 2

    @property
    def nv(self) -> int:
        """The number of virtual (unoccupied orbitals)."""
        return self.active - self.no

    @property
    def h_fermion_op(self) -> FermionOperator:
        """
        Hamiltonian as openfermion.FermionOperator
        """
        if self.hcb:
            raise ValueError("No FermionOperator available for hard-core boson Hamiltonian")
        return get_hop_from_integral(self.int1e, self.int2e) + self.e_core

    @property
    def h_qubit_op(self) -> QubitOperator:
        """
        Hamiltonian as openfermion.QubitOperator, mapped by
        Jordan-Wigner transformation.
        """
        if not self.hcb:
            return reverse_qop_idx(jordan_wigner(self.h_fermion_op), self.n_qubits)
        else:
            return get_hop_hcb_from_integral(self.int1e, self.int2e)

    @property
    def n_params(self) -> int:
        """The number of parameter in the ansatz/circuit."""
        # this definition ensures that `param[param_id]` is always valid
        if not self.param_ids:
            return 0
        return max(self.param_ids) + 1

    @property
    def statevector_size(self) -> int:
        """The size of the statevector."""
        return 1 << self.n_qubits

    @property
    def civector_size(self) -> int:
        """
        The size of the CI vector.
        """
        if not self.hcb:
            return round(comb(self.n_qubits // 2, self.n_elec // 2)) ** 2
        else:
            return round(comb(self.n_qubits, self.n_elec // 2))

    @property
    def init_state(self) -> Tensor:
        """
        The circuit initial state before applying the excitation operators. Usually RHF.

        See Also
        --------
        get_init_state_dataframe: Returns initial state information dataframe.
        """
        return self._init_state

    @init_state.setter
    def init_state(self, init_state):
        self._init_state = init_state

    @property
    def params(self) -> Tensor:
        """The circuit parameters."""
        if self._params is not None:
            return self._params
        if self.opt_res is not None:
            return self.opt_res.x
        return None

    @params.setter
    def params(self, params):
        self._params = params

    @property
    def param_ids(self) -> List[int]:
        """The mapping from excitations operators to parameters."""
        if self._param_ids is None:
            if self.ex_ops is None:
                raise ValueError("Excitation operators not defined")
            else:
                return tuple(range(len(self.ex_ops)))
        return self._param_ids

    @param_ids.setter
    def param_ids(self, v):
        self._param_ids = v



import numpy as np

import pytest
from scipy.sparse import linalg
from pyscf import fci
from openfermion.linalg import eigenspectrum
import tensorcircuit as tc

from tencirchem import UCCSD
from tencirchem.molecule import h4, h6, _random
from tencirchem.static.hamiltonian import get_h_from_hf, mpo_to_quoperator
from tencirchem.static.ci_utils import get_ci_strings
from tencirchem.static.hea import binary, parity


@pytest.mark.parametrize("m", [h4, _random(4, 4)])
@pytest.mark.parametrize("hcb", [False, True])
@pytest.mark.parametrize("htype", ["sparse"])
def test_hamiltonian(m, hcb, htype):
    hf = m.HF()
    hf.chkfile = None
    hf.verbose = 0
    hf.kernel()

    hamiltonian = get_h_from_hf(hf, hcb=hcb, htype=htype)
    if htype == "mpo":
        hamiltonian = mpo_to_quoperator(hamiltonian).eval_matrix()
    else:
        hamiltonian = np.array(hamiltonian.todense())

    e_nuc = hf.energy_nuc()
    if not hcb:
        fci_e, _ = fci.FCI(hf).kernel()
        # not generally true but works for this example
        np.testing.assert_allclose(np.linalg.eigh(hamiltonian)[0][0] + e_nuc, fci_e, atol=1e-6)
    else:
        circuit = tc.Circuit(4)
        for i in range(4 // 2):
            circuit.X(3 - i)
        state = circuit.state()
        np.testing.assert_allclose(state @ (hamiltonian @ state).reshape(-1) + e_nuc, hf.e_tot)


@pytest.mark.parametrize("m", [h4, _random(4, 4)])
@pytest.mark.parametrize("hcb", [False, True])
def test_hamiltonian_fcifunc(m, hcb):
    hf = m.HF()
    hf.chkfile = None
    hf.verbose = 0
    hf.kernel()

    htype = "fcifunc"
    hamiltonian = get_h_from_hf(hf, hcb=hcb, htype=htype)
    hamiltonian_ref = get_h_from_hf(hf, hcb=hcb, htype="sparse").todense()

    if not hcb:
        n_qubits = 8
    else:
        n_qubits = 4
    ci_strings = get_ci_strings(n_qubits, 4, hcb)
    hamiltonian = linalg.LinearOperator(shape=(len(ci_strings), len(ci_strings)), matvec=hamiltonian, dtype=np.float64)
    e = linalg.eigsh(hamiltonian, k=1, which="SA")[0][0]
    e_ref = np.linalg.eigh(hamiltonian_ref)[0][0]
    np.testing.assert_allclose(e, e_ref)


@pytest.mark.parametrize("m", [h4, _random(4, 4), h6])
@pytest.mark.parametrize("encode", [binary, parity])
def test_hea_encoding(m, encode):
    uccsd = UCCSD(m)
    qubit_hamiltonian = encode(uccsd.h_fermion_op, uccsd.n_qubits, uccsd.n_elec)
    e1 = eigenspectrum(qubit_hamiltonian)[0]
    np.testing.assert_allclose(e1, uccsd.e_fci)



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


import logging
from inspect import isfunction
from itertools import product
from typing import Tuple

import numpy as np
import tensornetwork as tn
from renormalizer import Mpo
from openfermion import FermionOperator, QubitOperator
from pyscf.gto.mole import Mole
from pyscf.scf.hf import RHF
from pyscf.mcscf import CASCI
from pyscf.fci import direct_nosym, cistring
from pyscf import ao2mo
import tensorcircuit as tc
from tensorcircuit import QuOperator

from tencirchem.utils.misc import hcb_to_coo, fop_to_coo, reverse_qop_idx, canonical_mo_coeff, get_n_qubits
from tencirchem.constants import DISCARD_EPS


logger = logging.getLogger(__name__)


def get_integral_from_hf(hf: RHF, active_space: Tuple = None):
    if not isinstance(hf, RHF):
        raise TypeError(f"hf object must be RHF class, got {type(hf)}")
    m = hf.mol
    assert hf.mo_coeff is not None
    # todo(weitangli): don't modify inplace
    hf.mo_coeff = canonical_mo_coeff(hf.mo_coeff)

    if active_space is None:
        nelecas = m.nelectron
        ncas = m.nao
    else:
        nelecas, ncas = active_space

    casci = CASCI(hf, ncas, nelecas)
    int1e, e_core = casci.get_h1eff()
    int2e = ao2mo.restore("s1", casci.get_h2eff(), ncas)

    return int1e, int2e, e_core


def get_hop_from_integral(int1e, int2e):
    n_orb = int1e.shape[0]
    if int1e.shape != (n_orb, n_orb):
        raise ValueError(f"Invalid one-boby integral array shape: {int1e.shape}")
    if int2e.shape != (n_orb, n_orb, n_orb, n_orb):
        raise ValueError(f"Invalid two-boby integral array shape: {int2e.shape}")
    n_sorb = n_orb * 2

    logger.info("Creating Hamiltonian operators")

    h1e = np.zeros((n_sorb, n_sorb))
    h2e = np.zeros((n_sorb, n_sorb, n_sorb, n_sorb))

    h1e[:n_orb, :n_orb] = h1e[n_orb:, n_orb:] = int1e

    for p, q, r, s in product(range(n_sorb), repeat=4):
        # a_p^\dagger a_q^\dagger a_r a_s
        if ((p < n_orb) == (s < n_orb)) and ((q < n_orb) == (r < n_orb)):
            # note the different orders of the indices
            h2e[p, q, r, s] = int2e[p % n_orb, s % n_orb, q % n_orb, r % n_orb]

    op1e = []
    for p, q in product(range(n_sorb), repeat=2):
        # a_p^\dagger a_q
        v = h1e[p, q]
        if np.abs(v) < DISCARD_EPS:
            continue
        op = FermionOperator(f"{p}^ {q}", v)
        op1e.append(op)

    op2e = []
    for q, s in product(range(n_sorb), repeat=2):
        for p, r in product(range(q), range(s)):
            # a_p^\dagger a_q^\dagger a_r a_s
            v = h2e[p, q, r, s] - h2e[q, p, r, s]
            if np.abs(v) < DISCARD_EPS:
                continue
            op = FermionOperator(f"{p}^ {q}^ {r} {s}", v)
            op2e.append(op)

    logger.info("Summing Hamiltonian operators")
    ops = FermionOperator()
    for op in op1e + op2e:
        ops += op

    return ops


def qubit_operator(string: str, coeff: float) -> QubitOperator:
    ret = coeff
    terms = string.split(" ")
    for term in terms:
        if term[-1] == "^":
            sign = -1
            term = term[:-1]
        else:
            sign = 1
        idx = int(term)
        ret *= (QubitOperator(f"X{idx}") + sign * 1j * QubitOperator(f"Y{idx}")) / 2
    return ret


def get_hop_hcb_from_integral(int1e, int2e):
    # Hard core boson Hamiltonian
    # https://arxiv.org/pdf/2002.00035.pdf
    n_orb = int1e.shape[0]
    qop = QubitOperator()
    for p in range(n_orb):
        for q in range(p + 1):
            if p == q:
                qop += qubit_operator(f"{p}^ {p}", 2 * int1e[p, p] + int2e[p, p, p, p])
            else:
                qop += qubit_operator(f"{p}^ {q}", int2e[p, q, p, q])
                qop += qubit_operator(f"{q}^ {p}", int2e[q, p, q, p])
                qop += qubit_operator(f"{p}^ {p} {q}^ {q}", 4 * int2e[p, p, q, q] - 2 * int2e[p, q, p, q])
    qop = reverse_qop_idx(qop, n_orb)
    return qop


def get_h_sparse_from_integral(int1e, int2e, hcb=False, do_log=False):
    if not hcb:
        ops = get_hop_from_integral(int1e, int2e)
    else:
        ops = get_hop_hcb_from_integral(int1e, int2e)
    if do_log:
        logger.info("Constructing sparse Hamiltonian")
    if not hcb:
        h_sparse = fop_to_coo(ops, n_qubits=2 * len(int1e))
    else:
        h_sparse = hcb_to_coo(ops, n_qubits=len(int1e))
    if do_log:
        logger.info("Sparse Hamiltonian constructed")
    return h_sparse


def get_h_fcifunc_from_integral(int1e, int2e, n_elec):
    n_orb = len(int1e)
    h2e = direct_nosym.absorb_h1e(int1e, int2e, n_orb, n_elec, 0.5)

    def fci_func(civector):
        civector = tc.backend.numpy(civector).astype(np.float64)
        civector = direct_nosym.contract_2e(h2e, civector, norb=n_orb, nelec=n_elec)
        return tc.backend.convert_to_tensor(civector).astype(tc.rdtypestr)

    return fci_func


def get_h_fcifunc_hcb_from_integral(int1e, int2e, n_elec):
    # todo: how about using https://github.com/pyscf/doci
    n_orb = len(int1e)
    ci_strings = cistring.make_strings(range(n_orb), n_elec // 2)

    def fci_func(civector):
        res = tc.backend.zeros(len(civector), dtype=tc.rdtypestr)
        for p in range(n_orb):
            for q in range(p + 1):
                if p == q:
                    bitmask = 1 << p
                    arraymask = (ci_strings & bitmask) == bitmask
                    res += (civector * arraymask) * (2 * int1e[p, p] + int2e[p, p, p, p])
                else:
                    bitmask = (1 << p) + (1 << q)
                    excitation = ci_strings ^ bitmask
                    addr = cistring.strs2addr(n_orb, n_elec // 2, excitation)
                    flip = ci_strings ^ (1 << p)
                    masked_flip = flip & bitmask
                    arraymask = (masked_flip == bitmask) | (masked_flip == 0)
                    res += civector[addr] * arraymask * int2e[p, q, p, q]
                    arraymask = (ci_strings & bitmask) == bitmask
                    res += (civector * arraymask) * (4 * int2e[p, p, q, q] - 2 * int2e[p, q, p, q])
        return res

    return fci_func


def get_h_from_integral(int1e, int2e, n_elec, hcb: bool, htype: str):
    if htype == "sparse":
        hamiltonian = get_h_sparse_from_integral(int1e, int2e, hcb=hcb, do_log=True)
    else:
        assert htype.lower() == "fcifunc"
        if not hcb:
            hamiltonian = get_h_fcifunc_from_integral(int1e, int2e, n_elec)
        else:
            hamiltonian = get_h_fcifunc_hcb_from_integral(int1e, int2e, n_elec)
    return hamiltonian


def get_h_from_hf(hf: RHF, active_space: Tuple = None, hcb: bool = False, htype="sparse"):
    if not isinstance(hf, RHF):
        raise TypeError(f"hf object must RHF class, got {type(hf)}")
    htype = htype.lower()
    if not htype in ["sparse", "mpo", "fcifunc"]:
        raise ValueError(f"htype must be 'sparse' or 'mpo', got '{htype}'")
    int1e, int2e, e_core = get_integral_from_hf(hf, active_space)
    if active_space is None:
        n_elec = hf.mol.nelectron
    else:
        n_elec = active_space[0]

    hamiltonian = get_h_from_integral(int1e, int2e, n_elec, hcb, htype)

    if active_space is None:
        return hamiltonian
    else:
        return hamiltonian, e_core


def mpo_to_quoperator(mpo: Mpo):
    array_list = [m.array for m in mpo]
    # squeeze dim 1 out
    assert len(array_list) >= 2
    a, b, c, d = array_list[0].shape
    array_list[0] = array_list[0].reshape(b, c, d)
    a, b, c, d = array_list[-1].shape
    array_list[-1] = array_list[-1].reshape(a, b, c)
    # convert MPO to tensor-network
    node_list = [tn.Node(array) for array in array_list]

    node_list[0][2] ^ node_list[1][0]
    for i in range(1, len(node_list) - 1):
        node_list[i][3] ^ node_list[i + 1][0]

    in_edges = [node_list[0][0]] + [node[1] for node in node_list[1:]]
    out_edges = [node_list[0][1]] + [node[2] for node in node_list[1:]]

    qop = QuOperator(out_edges, in_edges)
    return qop


def apply_op(op, state):
    if isinstance(op, list):
        # in MPO form
        n_qubit = get_n_qubits(state)
        n_qubit_op = get_n_qubits(op)
        assert n_qubit == n_qubit_op
        mps = tc.QuVector.from_tensor(state.reshape([2] * n_qubit))
        h_qop = mpo_to_quoperator(op)
        return (h_qop @ mps).eval().reshape(-1)
    if isfunction(op):
        return op(state)
    else:
        return op @ state


def random_integral(nao: int, seed: int = 2077):
    np.random.seed(seed)
    int1e = np.random.uniform(-1, 1, size=(nao, nao))
    int2e = np.random.uniform(-1, 1, size=(nao, nao, nao, nao))
    int1e = 0.5 * (int1e + int1e.T)
    int2e = symmetrize_int2e(int2e)
    return int1e, int2e


def symmetrize_int2e(int2e):
    int2e = 0.25 * (
        int2e + int2e.transpose((0, 1, 3, 2)) + int2e.transpose((1, 0, 2, 3)) + int2e.transpose((2, 3, 0, 1))
    )
    int2e = 0.5 * (int2e + int2e.transpose(3, 2, 1, 0))
    return int2e



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.

import numpy as np

a = np.array([[0, 1], [0, 0]])
# a^\dagger
ad = a.T
# a^\dagger_i a_j
ad_a = np.kron(ad, a)
# a^\dagger_i a_j - a_j a^\dagger_i
ad_a_hc = ad_a - ad_a.T
adad_aa = np.kron(np.kron(np.kron(ad, ad), a), a)
adad_aa_hc = adad_aa - adad_aa.T
ad_a_hc2 = ad_a_hc @ ad_a_hc
adad_aa_hc2 = adad_aa_hc @ adad_aa_hc

DISCARD_EPS = 1e-12



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from typing import Tuple, List

import numpy as np
from pyscf.gto.mole import Mole

from tencirchem.static.ucc import UCC
from tencirchem.constants import DISCARD_EPS


class UCCSD(UCC):
    """
    Run UCCSD calculation. For a comprehensive tutorial see :doc:`/tutorial_jupyter/ucc_functions`.

    Examples
    --------
    >>> import numpy as np
    >>> from tencirchem import UCCSD
    >>> from tencirchem.molecule import h2
    >>> uccsd = UCCSD(h2)
    >>> e_ucc = uccsd.kernel()
    >>> np.testing.assert_allclose(e_ucc, uccsd.e_fci, atol=1e-10)
    >>> e_hf = uccsd.energy(np.zeros(uccsd.n_params))
    >>> np.testing.assert_allclose(e_hf, uccsd.e_hf, atol=1e-10)
    """

    def __init__(
        self,
        mol: Mole,
        init_method: str = "mp2",
        active_space: Tuple[int, int] = None,
        mo_coeff: np.ndarray = None,
        pick_ex2: bool = True,
        epsilon: float = DISCARD_EPS,
        sort_ex2: bool = True,
        engine: str = None,
        run_hf: bool = True,
        run_mp2: bool = True,
        run_ccsd: bool = True,
        run_fci: bool = True,
    ):
        r"""
        Initialize the class with molecular input.

        Parameters
        ----------
        mol: Mole
            The molecule as PySCF ``Mole`` object.
        init_method: str, optional
            How to determine the initial amplitude guess. Accepts ``"mp2"`` (default), ``"ccsd"``
            and ``"zeros"``.
        active_space: Tuple[int, int], optional
            Active space approximation. The first integer is the number of electrons and the second integer is
            the number or spatial-orbitals. Defaults to None.
        mo_coeff: np.ndarray, optional
            Molecule coefficients. If provided then RHF is skipped.
            Can be used in combination with the ``init_state`` attribute.
            Defaults to None which means RHF orbitals are used.
        pick_ex2: bool, optional
            Whether screen out two body excitations based on the inital guess amplitude.
            Defaults to True, which means excitations with amplitude less than ``epsilon`` (see below) are discarded.
        epsilon: float, optional
            The threshold to discard two body excitations. Defaults to 1e-12.
        sort_ex2: bool, optional
            Whether sort two-body excitations in the ansatz based on the initial guess amplitude.
            Large excitations come first. Defaults to True.
            Note this could lead to different ansatz for the same molecule at different geometry.
        engine: str, optional
            The engine to run the calculation. See :ref:`advanced:Engines` for details.
        run_hf: bool, optional
            Whether run HF for molecule orbitals. Defaults to ``True``.
        run_mp2: bool, optional
            Whether run MP2 for initial guess and energy reference. Defaults to ``True``.
        run_ccsd: bool, optional
            Whether run CCSD for initial guess and energy reference. Defaults to ``True``.
        run_fci: bool, optional
            Whether run FCI  for energy reference. Defaults to ``True``.

        See Also
        --------
        tencirchem.KUPCCGSD
        tencirchem.PUCCD
        tencirchem.UCC
        """
        super().__init__(
            mol,
            init_method,
            active_space,
            mo_coeff,
            engine=engine,
            run_hf=run_hf,
            run_mp2=run_mp2,
            run_ccsd=run_ccsd,
            run_fci=run_fci,
        )
        self.pick_ex2 = pick_ex2
        self.sort_ex2 = sort_ex2
        # screen out excitation operators based on t2 amplitude
        self.t2_discard_eps = epsilon
        self.ex_ops, self.param_ids, self.init_guess = self.get_ex_ops(self.t1, self.t2)

    def get_ex_ops(self, t1: np.ndarray = None, t2: np.ndarray = None) -> Tuple[List[Tuple], List[int], List[float]]:
        """
        Get one-body and two-body excitation operators for UCCSD ansatz.
        Pick and sort two-body operators if ``self.pick_ex2`` and ``self.sort_ex2`` are set to ``True``.

        Parameters
        ----------
        t1: np.ndarray, optional
            Initial one-body amplitudes based on e.g. CCSD
        t2: np.ndarray, optional
            Initial two-body amplitudes based on e.g. MP2

        Returns
        -------
        ex_op: List[Tuple]
            The excitation operators. Each operator is represented by a tuple of ints.
        param_ids: List[int]
            The mapping from excitations to parameters.
        init_guess: List[float]
            The initial guess for the parameters.

        See Also
        --------
        get_ex1_ops: Get one-body excitation operators.
        get_ex2_ops: Get two-body excitation operators.

        Examples
        --------
        >>> from tencirchem import UCCSD
        >>> from tencirchem.molecule import h2
        >>> uccsd = UCCSD(h2)
        >>> ex_op, param_ids, init_guess = uccsd.get_ex_ops()
        >>> ex_op
        [(3, 2), (1, 0), (1, 3, 2, 0)]
        >>> param_ids
        [0, 0, 1]
        >>> init_guess  # doctest:+ELLIPSIS
        [0.0, ...]
        """
        ex1_ops, ex1_param_ids, ex1_init_guess = self.get_ex1_ops(self.t1)
        ex2_ops, ex2_param_ids, ex2_init_guess = self.get_ex2_ops(self.t2)

        # screen out symmetrically not allowed excitation
        ex2_ops, ex2_param_ids, ex2_init_guess = self.pick_and_sort(
            ex2_ops, ex2_param_ids, ex2_init_guess, self.pick_ex2, self.sort_ex2
        )

        ex_op = ex1_ops + ex2_ops
        param_ids = ex1_param_ids + [i + max(ex1_param_ids) + 1 for i in ex2_param_ids]
        init_guess = ex1_init_guess + ex2_init_guess
        return ex_op, param_ids, init_guess

    def pick_and_sort(self, ex_ops, param_ids, init_guess, do_pick=True, do_sort=True):
        # sort operators according to amplitude
        if do_sort:
            sorted_ex_ops = sorted(zip(ex_ops, param_ids), key=lambda x: -np.abs(init_guess[x[1]]))
        else:
            sorted_ex_ops = list(zip(ex_ops, param_ids))
        ret_ex_ops = []
        ret_param_ids = []
        for ex_op, param_id in sorted_ex_ops:
            # discard operators with tiny amplitude.
            # The default eps is so small that the screened out excitations are probably not allowed
            if do_pick and np.abs(init_guess[param_id]) < self.t2_discard_eps:
                continue
            ret_ex_ops.append(ex_op)
            ret_param_ids.append(param_id)
        unique_ids = np.unique(ret_param_ids)
        ret_init_guess = np.array(init_guess)[unique_ids]
        id_mapping = {old: new for new, old in enumerate(unique_ids)}
        ret_param_ids = [id_mapping[i] for i in ret_param_ids]
        return ret_ex_ops, ret_param_ids, list(ret_init_guess)

    @property
    def e_uccsd(self) -> float:
        """
        Returns UCCSD energy
        """
        return self.energy()



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from functools import partial

import numpy as np
from pyscf.fci import cistring
import tensorcircuit as tc

from tencirchem.utils.backend import jit, tensor_set_elem, get_xp, get_uint_type


def get_ci_strings(n_qubits, n_elec, hcb, strs2addr=False):
    xp = get_xp(tc.backend)
    uint_type = get_uint_type()
    if 2**n_qubits > np.iinfo(uint_type).max:
        raise ValueError(f"Too many qubits: {n_qubits}, try using complex128 datatype")
    if not hcb:
        beta = cistring.make_strings(range(n_qubits // 2), n_elec // 2)
        beta = xp.array(beta, dtype=uint_type)
        alpha = beta << (n_qubits // 2)
        ci_strings = (alpha.reshape(-1, 1) + beta.reshape(1, -1)).ravel()
        if strs2addr:
            strs2addr = xp.zeros(2 ** (n_qubits // 2), dtype=uint_type)
            strs2addr[beta] = xp.arange(len(beta))
            return ci_strings, strs2addr
    else:
        ci_strings = cistring.make_strings(range(n_qubits), n_elec // 2).astype(uint_type)
        if strs2addr:
            strs2addr = xp.zeros(2**n_qubits, dtype=uint_type)
            strs2addr[ci_strings] = xp.arange(len(ci_strings))
            return ci_strings, strs2addr

    return ci_strings


def get_addr(excitation, n_qubits, n_elec, strs2addr, hcb, num_strings=None):
    if hcb:
        return strs2addr[excitation]
    alpha = excitation >> (n_qubits // 2)
    beta = excitation & (2 ** (n_qubits // 2) - 1)
    alpha_addr2 = strs2addr[alpha]
    beta_addr2 = strs2addr[beta]
    if num_strings is None:
        num_strings = cistring.num_strings(n_qubits // 2, n_elec // 2)
    return alpha_addr2 * num_strings + beta_addr2


def get_ex_bitstring(n_qubits, n_elec, ex_op, hcb):
    if not hcb:
        bitstring_base = ["0"] * (n_qubits // 2 - n_elec // 2) + ["1"] * (n_elec // 2)
        bitstring_base *= 2
    else:
        bitstring_base = ["0"] * (n_qubits - n_elec // 2) + ["1"] * (n_elec // 2)

    bitstring = bitstring_base.copy()[::-1]
    # first annihilation then creation
    if len(ex_op) == 2:
        bitstring[ex_op[1]] = "0"
        bitstring[ex_op[0]] = "1"
    else:
        assert len(ex_op) == 4
        bitstring[ex_op[3]] = "0"
        bitstring[ex_op[2]] = "0"
        bitstring[ex_op[1]] = "1"
        bitstring[ex_op[0]] = "1"

    return "".join(reversed(bitstring))


def civector_to_statevector(civector, n_qubits, ci_strings):
    statevector = tc.backend.zeros(2**n_qubits, dtype=tc.rdtypestr)
    return tensor_set_elem(statevector, ci_strings, civector)


def statevector_to_civector(statevector, ci_strings):
    return statevector[ci_strings]


@partial(jit, static_argnums=[0])
def get_init_civector(len_ci):
    civector = tc.backend.zeros(len_ci, dtype=tc.rdtypestr)
    civector = tensor_set_elem(civector, 0, 1)
    return civector



import numpy as np
from pyscf import fci
from pyscf.scf import RHF

from tencirchem.molecule import _Molecule, h2o
from tencirchem.static.hamiltonian import get_integral_from_hf


def test_molecule():
    mol = h2o()
    mol.verbose = 0

    mf1 = RHF(mol)
    mf1.verbose = 0
    e_hf1 = mf1.kernel()
    e_fci1, _ = fci.FCI(mf1).kernel()

    int1e = mol.intor("int1e_kin") + mol.intor("int1e_nuc")
    int2e = mol.intor("int2e")
    n_elec = mol.nelectron
    e_nuc = mol.energy_nuc()
    ovlp = mol.intor("int1e_ovlp")
    mol = _Molecule(int1e, int2e, n_elec, e_nuc, ovlp)
    mf2 = RHF(mol)

    mf2.chkfile = None
    mf2.init_guess = "1e"
    mf2.verbose = 0
    e_hf2 = mf2.kernel()
    e_fci2, _ = fci.FCI(mf2).kernel()

    np.testing.assert_allclose(mf1.get_hcore(), mf2.get_hcore())
    np.testing.assert_allclose(mf1.get_ovlp(), mf2.get_ovlp())
    np.testing.assert_allclose(mf1._eri, mf2._eri, atol=1e-10)
    np.testing.assert_allclose(mf1.energy_nuc(), mf2.energy_nuc())
    np.testing.assert_allclose(mf1.energy_elec(), mf2.energy_elec())
    np.testing.assert_allclose(e_hf1, e_hf2)
    np.testing.assert_allclose(e_fci1, e_fci2)

    int1e, int2e, e_core = get_integral_from_hf(mf2)
    e_fci3, _ = fci.direct_spin1.kernel(int1e, int2e, mol.nao, mol.n_elec)
    np.testing.assert_allclose(e_fci2, e_fci3 + e_core)



import numpy as np
from pyscf import fci
from pyscf.scf import RHF
from pyscf.mcscf import CASCI
import pytest

from tencirchem import UCCSD, KUPCCGSD
from tencirchem.static.hamiltonian import get_integral_from_hf, random_integral
from tencirchem.molecule import _random, h4, h_chain
from tencirchem.utils.misc import canonical_mo_coeff


def get_random_integral_and_fci(n):
    nao = n_elec = n
    int1e, int2e = random_integral(nao)
    e, _ = fci.direct_spin1.kernel(int1e, int2e, nao, n_elec)
    return int1e, int2e, e


@pytest.mark.parametrize("hamiltonian", ["H4", "H4 integral", "random integral"])
@pytest.mark.parametrize("ansatz_str", ["UCCSD", "kUpCCGSD"])
def test_ucc(hamiltonian, ansatz_str):
    m = h4
    nao = m.nao
    n_elec = m.nelectron
    if ansatz_str == "UCCSD":
        ansatz = UCCSD
        kwargs = {}
        atol = 1e-4
        if hamiltonian == "random integral":
            atol = 1e-3
    elif ansatz_str == "kUpCCGSD":
        ansatz = KUPCCGSD
        kwargs = {"n_tries": 1}
        # too few tries and the error could be large
        atol = 3e-3
    else:
        assert False
    if hamiltonian == "H4":
        # from mol
        ucc = ansatz(m, **kwargs)
    elif hamiltonian == "H4 integral":
        int1e = m.intor("int1e_kin") + m.intor("int1e_nuc")
        int2e = m.intor("int2e")
        ovlp = m.intor("int1e_ovlp")
        n_elec = m.nelectron
        e_nuc = m.energy_nuc()
        ucc = ansatz.from_integral(int1e, int2e, n_elec, e_nuc, ovlp, **kwargs)
    else:
        int1e, int2e, _ = get_random_integral_and_fci(nao)
        ucc = ansatz.from_integral(int1e, int2e, n_elec, **kwargs)
    e = ucc.kernel()
    np.testing.assert_allclose(e, ucc.e_fci, atol=atol)


def test_rdm():
    m = _random(4, 4)
    hf = RHF(m)
    hf.chkfile = None
    hf.kernel()
    hf.mo_coeff = canonical_mo_coeff(hf.mo_coeff)
    my_fci = fci.FCI(hf)
    e1, fcivec = my_fci.kernel()
    # rdm in MO basis
    rdm1, rdm2 = fci.direct_spin1.make_rdm12(fcivec, 4, 4)
    uccsd = UCCSD(m)
    e2 = uccsd.kernel()
    np.testing.assert_allclose(e1, e2, atol=1e-3)
    rdm1_uccsd = uccsd.make_rdm1(basis="MO")
    rdm2_uccsd = uccsd.make_rdm2(basis="MO")
    np.testing.assert_allclose(rdm1_uccsd, rdm1, atol=5e-3)
    np.testing.assert_allclose(rdm2_uccsd, rdm2, atol=5e-3)

    int1e, int2e, e_core = get_integral_from_hf(hf)
    rdm_e = int1e.ravel() @ rdm1_uccsd.ravel() + 1 / 2 * int2e.ravel() @ rdm2_uccsd.ravel() + e_core
    np.testing.assert_allclose(e2, rdm_e, atol=1e-5)


def test_active_space():
    m = h_chain(12)
    m.verbose = 0
    ncas = 4
    nelecas = 2

    hf = m.HF()
    hf.kernel()
    hf.mo_coeff = canonical_mo_coeff(hf.mo_coeff)
    casci = CASCI(hf, ncas, nelecas)
    e1 = casci.kernel()[0]
    uccsd = UCCSD(m, active_space=(nelecas, ncas))
    e2 = uccsd.kernel()
    np.testing.assert_allclose(e1, e2, atol=1e-5)
    np.testing.assert_allclose(uccsd.make_rdm1(), casci.make_rdm1(), atol=1e-3)
    from pyscf.mcscf.addons import make_rdm12

    _, rdm2 = make_rdm12(casci)
    np.testing.assert_allclose(uccsd.make_rdm2(), rdm2, atol=1e-3)

    uccsd.print_summary(include_circuit=True)


def test_get_circuit():
    uccsd = UCCSD(h4)
    params = np.random.rand(uccsd.n_params)
    s1 = uccsd.get_circuit(params).state()
    s2 = uccsd.get_circuit(params, decompose_multicontrol=True).state()
    np.testing.assert_allclose(s2, s1, atol=1e-10)
    s3 = uccsd.get_circuit(params, trotter=True).state()
    np.testing.assert_allclose(s3, s1, atol=1e-10)



import numpy as np
import pytest
from qiskit.circuit.library import RealAmplitudes

from tencirchem import UCCSD, HEA, parity
from tencirchem.molecule import h2
from tests.static.test_engine import set_backend_with_skip


@pytest.mark.parametrize(
    "engine",
    [
        "tensornetwork",
        "tensornetwork-noise",
        "tensornetwork-shot",
        "tensornetwork-noise&shot",
    ],
)
@pytest.mark.parametrize("backend_str", ["jax", "numpy"])
@pytest.mark.parametrize("grad", ["param-shift", "autodiff", "free"])
def test_hea(engine, backend_str, grad, reset_backend):
    if backend_str in ["numpy", "cupy"] and grad == "autodiff":
        pytest.xfail("Incompatible backend and gradient method")
    if engine in ["tensornetwork-shot", "tensornetwork-noise&shot"] and grad == "autodiff":
        pytest.xfail("Incompatible engine and gradient method")
    set_backend_with_skip(backend_str)
    m = h2
    uccsd = UCCSD(m)
    hea = HEA.ry(uccsd.int1e, uccsd.int2e, uccsd.n_elec, uccsd.e_core, 3, engine=engine)
    hea.grad = grad
    e = hea.kernel()
    atol = 0.1
    if engine == "tensornetwork-noise&shot" and grad == "free":
        atol *= 2
    np.testing.assert_allclose(e, uccsd.e_fci, atol=atol)


def test_qiskit_circuit():
    m = h2
    uccsd = UCCSD(m)
    circuit = RealAmplitudes(2)
    hea = HEA(parity(uccsd.h_fermion_op, 4, 2), circuit, np.random.rand(circuit.num_parameters))
    e = hea.kernel()
    np.testing.assert_allclose(e, uccsd.e_fci, atol=1e-5)
    hea.print_summary()


@pytest.mark.parametrize("mapping", ["jordan-wigner", "bravyi-kitaev"])
def test_mapping(mapping):
    uccsd= UCCSD(h2)
    hea = HEA.ry(uccsd.int1e, uccsd.int2e, uccsd.n_elec, uccsd.e_core, 3, mapping=mapping)
    e = hea.kernel()
    np.testing.assert_allclose(e, uccsd.e_fci)



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from functools import partial

import numpy as np
import tensorcircuit as tc
from tensorcircuit import Circuit, DMCircuit
from tensorcircuit.noisemodel import circuit_with_noise
from tensorcircuit.experimental import parameter_shift_grad
from tensorcircuit.cloud.wrapper import batch_expectation_ps

from tencirchem.utils.backend import jit


@partial(jit, static_argnums=[1])
def get_statevector(params, get_circuit):
    return get_circuit(params).state()


@partial(jit, static_argnums=[1, 2])
def get_densitymatrix(params, get_dmcircuit, noise_conf):
    if noise_conf is None:
        raise ValueError("NoiseConf not provided")
    circuit = get_dmcircuit(params)
    circuit = circuit_with_noise(circuit, noise_conf)
    return circuit.densitymatrix()


@partial(jit, static_argnums=[2])
def get_energy_tensornetwork(params, h_array, get_circuit):
    s = get_statevector(params, get_circuit)
    return (s.conj() @ (h_array @ s)).real


@partial(jit, static_argnums=[2, 3])
def get_energy_tensornetwork_noise(params, h_array, get_dmcircuit, noise_conf):
    dm = get_densitymatrix(params, get_dmcircuit, noise_conf)
    return tc.backend.trace(dm @ h_array).real


def sample_expectation_pauli(c, paulis, coeffs, shots, noise_conf):
    expectations = []
    for pauli, coef in zip(paulis, coeffs):
        x = []
        y = []
        z = []
        m = {"X": x, "Y": y, "Z": z}
        for idx, symbol in pauli:
            m[symbol].append(idx)
        expectation = coef
        if len(x + y + z) != 0:
            expectation *= c.sample_expectation_ps(x=x, y=y, z=z, shots=shots, noise_conf=noise_conf)
        expectations.append(expectation)
    # already real
    return sum(expectations)


@partial(jit, static_argnums=[1, 3, 4])
def get_energy_tensornetwork_shot(params, paulis, coeffs, get_circuit, shots):
    c = get_circuit(params)
    return sample_expectation_pauli(c, paulis, coeffs, shots, None)


@partial(jit, static_argnums=[1, 3, 4, 5])
def get_energy_tensornetwork_noise_shot(params, paulis, coeffs, get_dmcircuit, noise_conf, shots: int):
    dm = get_densitymatrix(params, get_dmcircuit, noise_conf)
    n_qubits = round(np.log2(dm.shape[0]))
    c = DMCircuit(n_qubits, dminputs=dm)
    return sample_expectation_pauli(c, paulis, coeffs, shots, noise_conf)


def get_energy_qpu(params, paulis, coeffs, get_circuit, shots: int):
    c: Circuit = get_circuit(params)
    pss = []
    symbol_mapping = {"X": 1, "Y": 2, "Z": 3}
    for pauli in paulis:
        ps = [0] * c.circuit_param["nqubits"]
        for idx, symbol in pauli:
            ps[idx] = symbol_mapping[symbol]
        pss.append(ps)
    es = []
    for _ in range((shots - 1) // 8192 + 1):
        e = batch_expectation_ps(c, pss, device="tianxuan_s1", ws=coeffs, shots=8192)
        es.append(e)
    print(es)
    return np.mean(es)


def _get_energy_and_grad(partial_get_energy, params, grad):
    if grad == "param-shift":
        e = partial_get_energy(params)
        grad_f = parameter_shift_grad(partial_get_energy)
        g = grad_f(params)
    else:
        assert grad == "autodiff"
        e, g = tc.backend.value_and_grad(partial_get_energy)(params)
    return e, g


@partial(jit, static_argnums=[2, 3])
def get_energy_and_grad_tensornetwork(params, h_array, get_circuit, grad):
    partial_get_energy = partial(get_energy_tensornetwork, h_array=h_array, get_circuit=get_circuit)
    return _get_energy_and_grad(partial_get_energy, params, grad)


@partial(jit, static_argnums=[2, 3, 4])
def get_energy_and_grad_tensornetwork_noise(params, h_array, get_dmcircuit, noise_conf, grad):
    partial_get_energy = partial(
        get_energy_tensornetwork_noise, h_array=h_array, get_dmcircuit=get_dmcircuit, noise_conf=noise_conf
    )
    return _get_energy_and_grad(partial_get_energy, params, grad)


@partial(jit, static_argnums=[1, 3, 4, 5])
def get_energy_and_grad_tensornetwork_shot(params, paulis, coeffs, get_circuit, shots, grad):
    partial_get_energy = partial(
        get_energy_tensornetwork_shot,
        paulis=paulis,
        coeffs=coeffs,
        get_circuit=get_circuit,
        shots=shots,
    )
    return _get_energy_and_grad(partial_get_energy, params, grad)


@partial(jit, static_argnums=[1, 3, 4, 5, 6])
def get_energy_and_grad_tensornetwork_noise_shot(params, paulis, coeffs, get_dmcircuit, noise_conf, shots, grad):
    partial_get_energy = partial(
        get_energy_tensornetwork_noise_shot,
        paulis=paulis,
        coeffs=coeffs,
        get_dmcircuit=get_dmcircuit,
        noise_conf=noise_conf,
        shots=shots,
    )
    return _get_energy_and_grad(partial_get_energy, params, grad)


def get_energy_and_grad_qpu(params, paulis, coeffs, get_circuit, shots: int, grad):
    partial_get_energy = partial(
        get_energy_qpu,
        paulis=paulis,
        coeffs=coeffs,
        get_circuit=get_circuit,
        shots=shots,
    )
    return _get_energy_and_grad(partial_get_energy, params, grad)



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from functools import wraps
from inspect import isfunction
from typing import List, Tuple

import numpy as np
from scipy.sparse import coo_matrix
from renormalizer import Model, Mpo, Op
from renormalizer.model.basis import BasisSet
from openfermion import FermionOperator, QubitOperator, jordan_wigner, get_sparse_operator
from openfermion.utils import hermitian_conjugated
from qiskit.quantum_info import SparsePauliOp
import tensorcircuit as tc

from tencirchem.constants import DISCARD_EPS


def csc_to_coo(csc):
    coo = coo_matrix(csc)
    mask = DISCARD_EPS < np.abs(coo.data.real)
    indices = np.array([coo.row[mask], coo.col[mask]]).T
    values = coo.data.real[mask].astype(tc.rdtypestr)
    return tc.backend.coo_sparse_matrix(indices=indices, values=values, shape=coo.shape)


def fop_to_coo(fop: FermionOperator, n_qubits: int, real: bool = True):
    op = get_sparse_operator(jordan_wigner(reverse_fop_idx(fop, n_qubits)), n_qubits=n_qubits)
    if real:
        op = op.real
    return csc_to_coo(op)


def hcb_to_coo(qop: QubitOperator, n_qubits: int, real: bool = True):
    op = get_sparse_operator(qop, n_qubits)
    if real:
        op = op.real
    return csc_to_coo(op)


def qop_to_qiskit(qop: QubitOperator, n_qubits: int) -> SparsePauliOp:
    sparse_list = []
    for k, v in qop.terms.items():
        s = "".join(kk[1] for kk in k)
        idx = [kk[0] for kk in k]
        sparse_list.append([s, idx, v])
    return SparsePauliOp.from_sparse_list(sparse_list, num_qubits=n_qubits)


def reverse_qop_idx(op: QubitOperator, n_qubits: int) -> QubitOperator:
    ret = QubitOperator()
    for pauli_string, v in op.terms.items():
        # internally QubitOperator assumes ascending index
        pauli_string = tuple(reversed([(n_qubits - 1 - idx, symbol) for idx, symbol in pauli_string]))
        ret.terms[pauli_string] = v
    return ret


def reverse_fop_idx(op: FermionOperator, n_qubits: int) -> FermionOperator:
    ret = FermionOperator()
    for word, v in op.terms.items():
        word = tuple([(n_qubits - 1 - idx, symbol) for idx, symbol in word])
        ret.terms[word] = v
    return ret


def format_ex_op(ex_op: Tuple) -> str:
    if len(ex_op) == 2:
        return f"{ex_op[0]}^ {ex_op[1]}"
    else:
        assert len(ex_op) == 4
        return f"{ex_op[0]}^ {ex_op[1]}^ {ex_op[2]} {ex_op[3]}"


def scipy_opt_wrap(f, gradient=True):
    @wraps(f)
    def _wrap_scipy_opt(_params, *args):
        # scipy assumes 64bit https://github.com/scipy/scipy/issues/5832
        res = f(tc.backend.convert_to_tensor(_params), *args)
        if gradient:
            return [np.asarray(tc.backend.numpy(v), dtype=np.float64) for v in res]
        else:
            return np.asarray(tc.backend.numpy(res), dtype=np.float64)

    return _wrap_scipy_opt


def rdm_mo2ao(rdm: np.ndarray, mo_coeff: np.ndarray):
    c = mo_coeff
    if rdm.ndim == 2:
        return c @ rdm @ c.T
    else:
        assert rdm.ndim == 4
        for _ in range(4):
            rdm = np.tensordot(rdm, c.T, axes=1).transpose(3, 0, 1, 2)
        return rdm


def canonical_mo_coeff(mo_coeff: np.ndarray):
    # make the first large element positive
    # all elements smaller than 1e-5 is highly unlikely (at least 1e10 basis)
    largest_elem_idx = np.argmax(1e-5 < np.abs(mo_coeff), axis=0)
    largest_elem = mo_coeff[(largest_elem_idx, np.arange(len(largest_elem_idx)))]
    return mo_coeff * np.sign(largest_elem).reshape(1, -1)


def get_n_qubits(vector_or_matrix_or_mpo_func):
    if isinstance(vector_or_matrix_or_mpo_func, list):
        return len(vector_or_matrix_or_mpo_func)
    if isfunction(vector_or_matrix_or_mpo_func):
        return vector_or_matrix_or_mpo_func.n_qubit
    # if hasattr(vector_or_matrix_or_mpo_func, "n_qubits"):
    #     return getattr(vector_or_matrix_or_mpo_func, "n_qubits")
    return round(np.log2(vector_or_matrix_or_mpo_func.shape[0]))


def ex_op_to_fop(ex_op, with_conjugation=False):
    if len(ex_op) == 2:
        fop = FermionOperator(f"{ex_op[0]}^ {ex_op[1]}")
    else:
        assert len(ex_op) == 4
        fop = FermionOperator(f"{ex_op[0]}^ {ex_op[1]}^ {ex_op[2]} {ex_op[3]}")
    if with_conjugation:
        fop = fop - hermitian_conjugated(fop)
    return fop


def get_dense_operator(basis: List[BasisSet], terms: List[Op]):
    return Mpo(Model(basis, []), terms).todense()



import numpy as np
import pytest
from renormalizer import Model, Mpo, Mps, Op, BasisHalfSpin, BasisSineDVR

from tencirchem import get_dense_operator, TimeEvolution, set_backend
from tencirchem.dynamic.model import pyrazine, sbm
from tencirchem.dynamic.transform import qubit_encode_op, qubit_encode_basis, transform_op, get_subspace_idx


def test_transform():
    op = Op("X Y", 0)
    basis = BasisHalfSpin(0)

    op_transformed = transform_op(op, basis)
    assert len(op_transformed) == 1
    assert op_transformed[0] == Op("Z", 0, 1j)

    op = Op("x", 0)
    basis = BasisSineDVR(0, 4, 0, 1)
    op_transformed = transform_op(op, basis, "binary")
    basis_transformed = qubit_encode_basis([basis], "binary")
    mat1 = get_dense_operator([basis], op)
    mat2 = get_dense_operator(basis_transformed, op_transformed)
    np.testing.assert_allclose(mat1, mat2, atol=1e-10)


@pytest.mark.parametrize("module", [pyrazine, sbm])
@pytest.mark.parametrize("nlevels", [2, 3])
@pytest.mark.parametrize("boson_encoding", [None, "unary", "binary", "gray"])
def test_qubit_encoding(module, nlevels, boson_encoding):
    if boson_encoding is None and nlevels == 3:
        pytest.skip()

    if module == pyrazine:
        nmode = 4
        ham_terms = pyrazine.get_ham_terms()
        basis = pyrazine.get_basis(nlevels)
    else:
        assert module == sbm
        epsilon = 0.1
        delta = 0.23
        nmode = 3
        omega_list = np.random.rand(nmode)
        g_list = np.random.rand(nmode)

        ham_terms = sbm.get_ham_terms(epsilon, delta, nmode, omega_list, g_list)
        basis = sbm.get_basis(omega_list, nlevels=nlevels)

    h1 = Mpo(Model(basis, ham_terms)).todense()

    spin_ham_terms, constant = qubit_encode_op(ham_terms, basis, boson_encoding=boson_encoding)
    # sth must be off if there are so many terms
    assert len(spin_ham_terms) < 200
    spin_basis = qubit_encode_basis(basis, boson_encoding)
    spin_ham_terms[-1] = complex(1) * spin_ham_terms[-1]
    h2 = Mpo(Model(spin_basis, spin_ham_terms)).todense().real
    h2 += np.eye(len(h2)) * constant

    subspace_idx = get_subspace_idx(basis, boson_encoding)
    np.testing.assert_allclose(h2[subspace_idx][:, subspace_idx], h1, atol=1e-10)


@pytest.mark.parametrize("boson_encoding", ["unary", "binary", "gray"])
def test_init_circuit(boson_encoding, reset_backend):
    backend = set_backend("jax")
    nlevels = 3
    ham_terms = pyrazine.get_ham_terms()
    basis = pyrazine.get_basis(nlevels)
    init_condition = {"10a": 2, "6a": [np.sqrt(2) / 2, 0, -np.sqrt(2) / 2]}
    tc = TimeEvolution(ham_terms, basis, boson_encoding=boson_encoding, init_condition=init_condition)
    ref_state = Mps.hartree_product_state(Model(basis, ham_terms), condition=init_condition).todense()

    subspace_idx = get_subspace_idx(basis, boson_encoding)

    np.testing.assert_allclose(backend.numpy(tc.state)[subspace_idx], ref_state)



__version__ = "2023.03"
__author__ = "TenCirChem Authors"
__creator__ = "Weitang Li"

#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.

import os
import logging

os.environ["JAX_ENABLE_X64"] = "True"
# for debugging
# os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

# disable CUDA 11.1 warning
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

logger = logging.getLogger("tensorcircuit")
logger.setLevel(logging.FATAL)

os.environ["RENO_LOG_LEVEL"] = "100"

logger = logging.getLogger("tencirchem")
logger.setLevel(logging.WARNING)

# finish logger stuff
del logger

from tencirchem.utils.backend import set_backend, set_dtype

# by default use float64 rather than float32
set_dtype("complex128")

# static module
# as an external interface
from pyscf import M

from tencirchem.static.ucc import UCC
from tencirchem.static.uccsd import UCCSD
from tencirchem.static.kupccgsd import KUPCCGSD
from tencirchem.static.puccd import PUCCD
from tencirchem.static.hea import HEA, parity, binary, get_noise_conf

# dynamic module
# as an external interface
from renormalizer import Op, BasisSHO, BasisHalfSpin, BasisSimpleElectron, BasisMultiElectron, Model, Mpo
from renormalizer.model import OpSum

from tencirchem.utils.misc import get_dense_operator
from tencirchem.dynamic.time_evolution import TimeEvolution


def clear_cache():
    from tencirchem.utils.backend import ALL_JIT_LIBS
    from .static.evolve_civector import CI_OPERATOR_CACHE, CI_OPERATOR_BATCH_CACHE

    for l in ALL_JIT_LIBS:
        l.clear()
    CI_OPERATOR_CACHE.clear()
    CI_OPERATOR_BATCH_CACHE.clear()



import numpy as np
import pytest

from tencirchem import UCCSD, set_backend
from tencirchem.molecule import h4
from tencirchem.static.engine_ucc import apply_excitation
from tencirchem.static.ci_utils import get_init_civector
from tencirchem.static.evolve_tensornetwork import get_init_circuit


@pytest.fixture
def ref_eg():
    uccsd = UCCSD(h4)

    np.random.seed(2077)
    params = np.random.rand(len(uccsd.init_guess)) - 0.5

    set_backend("jax")
    # the simplest method is the most reliable one
    e, g = uccsd.energy_and_grad(params, engine="tensornetwork")
    set_backend("numpy")
    return e, g


@pytest.mark.parametrize("engine", ["statevector", "civector", "civector-large", "pyscf"])
def test_gradient(ref_eg, engine, reset_backend):
    e_ref, g_ref = ref_eg

    uccsd = UCCSD(h4)

    np.random.seed(2077)
    params = np.random.rand(len(uccsd.init_guess)) - 0.5

    if engine == "statevector":
        set_backend("jax")
    else:
        set_backend("numpy")
    e, g = uccsd.energy_and_grad(params, engine=engine)

    np.testing.assert_allclose(e, e_ref, atol=1e-5)
    np.testing.assert_allclose(g, g_ref, atol=1e-5)


@pytest.fixture
def ref_state():
    uccsd = UCCSD(h4)

    np.random.seed(2077)
    params = np.random.rand(len(uccsd.init_guess)) - 0.5
    set_backend("jax")
    engine = "tensornetwork"
    state = uccsd.civector(params, engine=engine)
    state = apply_excitation(state, n_qubits=8, n_elec=4, ex_op=(4, 0, 3, 7), hcb=False, engine=engine)
    set_backend("numpy")
    return state


@pytest.mark.parametrize("engine", ["statevector", "civector", "civector-large", "pyscf"])
def test_excitation(ref_state, engine, reset_backend):
    uccsd = UCCSD(h4)

    np.random.seed(2077)
    params = np.random.rand(len(uccsd.init_guess)) - 0.5

    if engine == "statevector":
        set_backend("jax")
    else:
        set_backend("numpy")
    state = uccsd.civector(params, engine=engine)
    state = apply_excitation(state, n_qubits=8, n_elec=4, ex_op=(4, 0, 3, 7), hcb=False, engine=engine)
    np.testing.assert_allclose(state, ref_state, atol=1e-6)


def set_backend_with_skip(backend_str):
    if backend_str == "jax":
        # Detect GPU, https://github.com/google/jax/issues/971
        from jax.lib import xla_bridge

        if xla_bridge.get_backend().platform == "cpu":
            pytest.skip("skip JAX + CPU because of the slow compilation")
    elif backend_str == "cupy":
        try:
            import cupy
        except ImportError:
            pytest.skip("CuPy is not installed")
    return set_backend(backend_str)


@pytest.mark.parametrize("engine", ["tensornetwork", "statevector", "civector", "civector-large", "pyscf"])
@pytest.mark.parametrize("backend_str", ["jax", "numpy", "cupy"])
@pytest.mark.parametrize("init_state", [None, "civector", "circuit"])
def test_gradient_opt(backend_str, engine, init_state, reset_backend):
    if engine in ["tensornetwork", "statevector"] and backend_str in ["numpy", "cupy"]:
        pytest.xfail("Incompatible engine and backend")
    set_backend_with_skip(backend_str)
    uccsd = UCCSD(h4, engine=engine)
    # test initial condition. Has no effect
    if init_state == "civector":
        uccsd.init_state = get_init_civector(uccsd.civector_size)
    elif init_state == "circuit":
        uccsd.init_state = get_init_circuit(uccsd.n_qubits, uccsd.n_elec, uccsd.hcb)
    e = uccsd.kernel()
    np.testing.assert_allclose(e, uccsd.e_fci, atol=1e-4)



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from tencirchem.utils.misc import scipy_opt_wrap



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


import numpy as np
from pyscf.gto.mole import Mole
from pyscf import M, ao2mo

from tencirchem.static.hamiltonian import random_integral


# Duck-typing PySCF Mole object. Not supposed to be an external user-interface
class _Molecule(Mole):
    @classmethod
    def random(cls, nao, n_elec, seed=2077):
        int1e, int2e = random_integral(nao, seed)
        return cls(int1e, int2e, n_elec)

    def __init__(self, int1e, int2e, n_elec: int, e_nuc: float = 0, ovlp: np.ndarray = None):
        super().__init__()

        self.nao = len(int1e)
        self.int1e = int1e
        self.int2e = int2e
        self.int2e_s8 = ao2mo.restore(8, self.int2e, self.nao)
        # in PySCF m.nelec returns a tuple and m.nelectron returns an integer
        # So here use n_elec to indicate the difference
        self.n_elec = self.nelectron = n_elec
        self.e_nuc = e_nuc
        if ovlp is None:
            self.ovlp = np.eye(self.nao)
        else:
            self.ovlp = ovlp
        # self.symmetry = True
        # avoid sanity check
        self.verbose = 0
        self.build()
        self.incore_anyway = True

    def intor(self, intor, comp=None, hermi=0, aosym="s1", out=None, shls_slice=None, grids=None):
        if intor == "int1e_kin":
            return np.zeros_like(self.int1e)
        elif intor == "int1e_nuc":
            return self.int1e
        elif intor == "int1e_ovlp":
            return self.ovlp
        elif intor == "int2e":
            assert aosym in ["s8", "s1"]
            if aosym == "s1":
                return self.int2e
            else:
                return self.int2e_s8
        else:
            raise ValueError(f"Unsupported integral type: {intor}")

    intor_symmetric = intor

    def tot_electrons(self):
        return self.n_elec

    def energy_nuc(self):
        return self.e_nuc


# for testing/benchmarking

_random = _Molecule.random


def h_chain(n_h=4, bond_distance=0.8, charge=0):
    return M(atom=[["H", 0, 0, bond_distance * i] for i in range(n_h)], charge=charge)


def h_ring(n_h=4, radius=0.7):
    atoms = []
    for angle in np.linspace(0, 2 * np.pi, n_h + 1)[:-1]:
        atom = ["H", 0, np.cos(angle) * radius, np.sin(angle) * radius]
        atoms.append(atom)
    return M(atom=atoms)


H2 = h2 = h_chain(2, 0.741)

H3p = h3p = h_chain(3, charge=1)

H4 = h4 = h_chain()

H5p = h5p = h_chain(5, charge=1)

H6 = h6 = h_chain(6)

H8 = h8 = h_chain(8)


def water(bond_angle=104.45, bond_length=0.9584, basis="sto3g"):
    bond_angle = bond_angle / 180 * np.pi
    phi = bond_angle / 2
    r = bond_length
    O = ["O", 0, 0, 0]
    H1 = ["H", -r * np.sin(phi), r * np.cos(phi), 0]
    H2 = ["H", r * np.sin(phi), r * np.cos(phi), 0]
    return M(atom=[O, H1, H2], basis=basis)


H2O = h2o = water


HeHp = hehp = lambda d=1: M(atom=[["H", 0, 0, 0], ["He", 0, 0, d]], charge=1)
LiH = lih = lambda d=1.6: M(atom=[["H", 0, 0, 0], ["Li", 0, 0, d]])
BeH2 = beh2 = lambda d=1.6: M(atom=[["H", 0, 0, -d], ["Be", 0, 0, 0], ["H", 0, 0, d]])

N2 = n2 = nitrogen = lambda x: M(atom=[["N", 0, 0, 0], ["N", 0, 0, 1.09]])

ch4_coord = [
    ["C", 0, 0, 0],
    ["H", -0.629, 0.629, 0.629],
    ["H", 0.629, -0.629, 0.629],
    ["H", -0.629, -0.629, -0.629],
    ["H", 0.629, 0.629, -0.629],
]
CH4 = ch4 = methane = lambda x: M(atom=ch4_coord)


if __name__ == "__main__":
    n = 6
    h1 = np.zeros((n, n))
    for i in range(n - 1):
        h1[i, i + 1] = h1[i + 1, i] = -1.0
    h1[n - 1, 0] = h1[0, n - 1] = -1.0
    eri = np.zeros((n, n, n, n))
    for i in range(n):
        eri[i, i, i, i] = 2.0

    m = _Molecule(h1, eri, 6)
    from pyscf.scf import RHF

    rhf = RHF(m)
    # avoid serialization warning
    rhf.chkfile = False
    print(rhf.kernel())



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from typing import Tuple

import pandas as pd
from opt_einsum import contract_path
import tensornetwork as tn
import tensorcircuit as tc


def get_circuit_dataframe(circuit: tc.Circuit):
    n_gates = circuit.gate_count()
    n_cnot = circuit.gate_count(["cnot"])
    n_mc = circuit.gate_count(["multicontrol"])
    # avoid int overflow and for better formatting
    n_qubits = circuit.circuit_param["nqubits"]
    flop = count_circuit_flop(circuit)

    df_dict = {
        "#qubits": n_qubits,
        "#gates": [n_gates],
        "#CNOT": [n_cnot],
        "#multicontrol": [n_mc],
        "depth": [circuit.to_qiskit().depth()],
        "#FLOP": [flop],
    }
    return pd.DataFrame(df_dict)


def count_circuit_flop(circuit: tc.Circuit):
    # tensor network contraction flops by greedy algorithm
    nodes = circuit._copy()[0]
    input_set_list = [set([id(e) for e in node.edges]) for node in nodes]
    array_list = [node.tensor for node in nodes]
    output_set = set([id(e) for e in tn.get_subgraph_dangling(nodes)])
    args = []
    for i in range(len(nodes)):
        args.extend([array_list[i], input_set_list[i]])
    args.append(output_set)
    _, desc = contract_path(*args)
    return desc.opt_cost


def evolve_pauli(circuit: tc.Circuit, pauli_string: Tuple, theta: float):
    # pauli_string in openfermion.QubitOperator.terms format
    for idx, symbol in pauli_string:
        if symbol == "X":
            circuit.H(idx)
        elif symbol == "Y":
            circuit.SD(idx)
            circuit.H(idx)
        elif symbol == "Z":
            continue
        else:
            raise ValueError(f"Invalid Pauli String: {pauli_string}")

    for i in range(len(pauli_string) - 1):
        circuit.CNOT(pauli_string[i][0], pauli_string[i + 1][0])
    circuit.rz(pauli_string[-1][0], theta=theta)

    for i in reversed(range(len(pauli_string) - 1)):
        circuit.CNOT(pauli_string[i][0], pauli_string[i + 1][0])

    for idx, symbol in pauli_string:
        if symbol == "X":
            circuit.H(idx)
        elif symbol == "Y":
            circuit.H(idx)
            circuit.S(idx)
        elif symbol == "Z":
            continue
        else:
            raise ValueError(f"Invalid Pauli String: {pauli_string}")
    return circuit


def multicontrol_ry(theta):
    # https://arxiv.org/pdf/2005.14475.pdf
    c = tc.Circuit(4)
    i, j, k, l = 0, 1, 2, 3

    c.x(i)
    c.x(k)

    c.ry(l, theta=theta / 8)
    c.h(k)
    c.cnot(l, k)

    c.ry(l, theta=-theta / 8)
    c.h(i)
    c.cnot(l, i)

    c.ry(l, theta=theta / 8)
    c.cnot(l, k)

    c.ry(l, theta=-theta / 8)
    c.h(j)
    c.cnot(l, j)

    c.ry(l, theta=theta / 8)
    c.cnot(l, k)

    c.ry(l, theta=-theta / 8)
    c.cnot(l, i)

    c.ry(l, theta=theta / 8)
    c.h(i)
    c.cnot(l, k)

    # there's a typo in the paper
    c.ry(l, theta=-theta / 8)
    c.h(k)
    c.cnot(l, j)
    c.h(j)

    c.x(i)
    c.x(k)
    return c



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


import time
import logging
from functools import partial
from collections import defaultdict
from typing import Dict, List, Optional

import numpy as np
import scipy
from scipy.integrate import solve_ivp
from scipy.optimize import minimize
import tensorcircuit as tc
from renormalizer import Model, Mps, Mpo, BasisHalfSpin
from renormalizer.model.basis import BasisSet

from tencirchem.dynamic.transform import qubit_encode_op, qubit_encode_basis, get_init_circuit
from tencirchem.dynamic.time_derivative import (
    get_circuit,
    get_ansatz,
    get_jacobian_func,
    get_deriv,
    get_pvqd_loss_func,
    one_trotter_step,
)

logger = logging.getLogger(__name__)


class Ivp_Config:
    def __init__(self, method="RK45", rtol=1e-3, atol=1e-6):
        self.method = method
        self.rtol = rtol
        self.atol = atol


def evolve_exact(evals: np.ndarray, evecs: np.ndarray, init: np.ndarray, t: float):
    return evecs @ (np.diag(np.exp(-1j * t * evals)) @ (evecs.T @ init))


class TimeEvolution:
    def __init__(
        self,
        ham_terms,
        basis: List[BasisSet],
        boson_encoding: str = "gray",
        init_condition: Optional[Dict] = None,
        n_layers: int = 3,
        eps: float = 1e-5,
        property_op_dict: Dict = None,
        ref_only: bool = False,
        ivp_config: Ivp_Config = None,
    ):
        # handling defaults
        if init_condition is None:
            init_condition = {}
        if property_op_dict is None:
            property_op_dict = {}

        # setup refs
        self.model_ref = Model(basis, ham_terms)
        self.h_mpo_ref = Mpo(self.model_ref)
        self.h_ref = self.h_mpo_ref.todense()
        self.evals_ref, self.evecs_ref = scipy.linalg.eigh(self.h_ref)
        self.init_ref = Mps.hartree_product_state(self.model_ref, init_condition).todense()
        # only do ref in kernel. Could be useful for debugging
        self.ref_only = ref_only

        # setup transformed Hamiltonian
        ham_terms_spin, self.constant = qubit_encode_op(ham_terms, basis, boson_encoding)
        basis_spin: List[BasisHalfSpin] = qubit_encode_basis(basis, boson_encoding)
        # help perform some sanity checks
        self.model = Model(basis_spin, ham_terms_spin)
        self.h_mpo = Mpo(self.model)
        self.h = self.h_mpo.todense()

        # setup the ansatz
        self.init_circuit = get_init_circuit(self.model_ref, self.model, boson_encoding, init_condition)
        self.n_layers = n_layers
        self.n_params = self.n_layers * len(self.model.ham_terms)
        self.ansatz = get_ansatz(self.model.ham_terms, self.model.basis, self.n_layers, self.init_circuit)
        self.jacobian_func = get_jacobian_func(self.ansatz)

        # setup runtime components
        self.current_circuit = self.init_circuit
        self.eps = eps
        self.include_phase = False
        if ivp_config is None:
            self.ivp_config = Ivp_Config()
        else:
            self.ivp_config = ivp_config

        def scipy_deriv(t, _params):
            return get_deriv(self.ansatz, self.jacobian_func, _params, self.h, self.eps, self.include_phase)

        self.scipy_deriv = scipy_deriv

        self.pvqd_loss = get_pvqd_loss_func(self.ansatz)

        def solve_pvqd(_params, delta_t):
            hamiltonian = tc.backend.convert_to_tensor(self.h)
            loss = partial(self.pvqd_loss, params=_params, hamiltonian=hamiltonian, delta_t=delta_t)
            opt_res = minimize(loss, np.zeros_like(_params), jac=True, method="L-BFGS-B")
            return opt_res

        self.solve_pvqd = solve_pvqd

        self.property_op_dict = property_op_dict
        self.property_mat_dict = {}
        for k, op in self.property_op_dict.items():
            transformed_op, constant = qubit_encode_op(op, basis, boson_encoding)
            mat = Mpo(self.model, transformed_op).todense()
            mat += constant * np.eye(len(mat))
            mat_ref = Mpo(self.model_ref, op).todense()
            self.property_mat_dict[k] = mat, mat_ref

        # time evolution result
        self.t_list = [0]
        self.params_list = [np.zeros(self.n_params, dtype=np.float64)]
        state = self.init_circuit.state()
        self.state_list = [state]
        state_ref = self.init_ref
        self.state_ref_list = [state_ref]
        self.scipy_sol_list = []
        self._property_dict = defaultdict(list)
        self.update_property_dict(state, state_ref)

        self.wall_time_list = []

    def kernel(self, tau, algo="vqd"):
        # one step of time evolution
        if self.ref_only:
            return self.kernel_ref_only(tau)
        time0 = time.time()
        if algo == "vqd" or algo == "pvqd":
            if algo == "vqd":
                method, rtol, atol = self.ivp_config.method, self.ivp_config.rtol, self.ivp_config.atol
                scipy_sol = solve_ivp(
                    self.scipy_deriv, [self.t, self.t + tau], self.params, method=method, rtol=rtol, atol=atol
                )
                new_params = scipy_sol.y[:, -1]
            else:
                scipy_sol = self.solve_pvqd(self.params, tau)
                new_params = self.params + scipy_sol.x

            self.params_list.append(new_params)
            state = self.ansatz(self.params)
            self.scipy_sol_list.append(scipy_sol)
        else:
            assert algo == "trotter"
            self.current_circuit = one_trotter_step(
                self.model.ham_terms, self.model.basis, self.current_circuit, tau, inplace=True
            )
            shortcut = one_trotter_step(self.model.ham_terms, self.model.basis, self.state, tau)
            state = shortcut.state()

        time1 = time.time()
        self.t_list.append(self.t + tau)
        # t and params already updated
        self.state_list.append(state)
        state_ref = evolve_exact(self.evals_ref, self.evecs_ref, self.init_ref, self.t)
        self.state_ref_list.append(state_ref)
        # calculate properties
        self.update_property_dict(state, state_ref)

        self.wall_time_list.append(time1 - time0)

    def kernel_ref_only(self, tau):
        # Let's do code duplication to keep the source code simple
        self.t_list.append(self.t + tau)
        state_ref = evolve_exact(self.evals_ref, self.evecs_ref, self.init_ref, self.t)
        self.state_ref_list.append(state_ref)
        # calculate properties
        self.update_property_dict(None, state_ref)

    def update_property_dict(self, state, state_ref):
        for k, (mat, mat_ref) in self.property_mat_dict.items():
            if not self.ref_only:
                res = state.T.conj() @ (mat @ state), state_ref.T.conj() @ (mat_ref @ state_ref)
            else:
                res = state_ref.T.conj() @ (mat_ref @ state_ref)
            self._property_dict[k].append(res)

    def get_circuit(self, params, param_ids=None, compile_evolution=False):
        return get_circuit(
            self.ham_terms_spin,
            self.basis_spin,
            self.n_layers,
            self.init_circuit,
            params,
            param_ids,
            compile_evolution=compile_evolution,
        )

    @property
    def ham_terms_spin(self):
        return self.model.ham_terms

    @property
    def basis_spin(self):
        return self.model.basis

    @property
    def t(self):
        return self.t_list[-1]

    @property
    def t_array(self):
        return np.array(self.t_list)

    @property
    def params(self):
        return self.params_list[-1]

    @property
    def params_array(self):
        return np.array(self.params_list)

    @property
    def state(self):
        return self.state_list[-1]

    @property
    def state_array(self):
        return np.array(self.state_list)

    @property
    def state_ref(self):
        return self.state_ref_list[-1]

    @property
    def state_ref_array(self):
        return np.array(self.state_ref_list)

    @property
    def property_dict(self):
        return {k: np.array(v) for k, v in self._property_dict.items()}

    properties = property_dict

    @property
    def wall_time(self):
        return self.wall_time_list[-1]



import numpy as np
import pytest
from renormalizer import Op

from tencirchem import set_backend
from tencirchem.dynamic import sbm, qubit_encode_op, qubit_encode_basis, TimeEvolution


@pytest.mark.parametrize("algorithm", ["vanilla", "include_phase", "p-VQD", "trotter"])
def test_sbm(reset_backend, algorithm):
    set_backend("jax")
    epsilon = 0
    delta = 1
    nmode = 1
    omega_list = [1]
    g_list = [0.5]
    nbas = 8
    n_layers = 3

    ham_terms = sbm.get_ham_terms(epsilon, delta, nmode, omega_list, g_list)
    basis = sbm.get_basis(omega_list, [nbas] * nmode)
    ham_terms_spin, _ = qubit_encode_op(ham_terms, basis, "gray")
    basis_spin = qubit_encode_basis(basis, "gray")

    te = TimeEvolution(
        ham_terms_spin,
        basis_spin,
        n_layers=n_layers,
        property_op_dict={"Z": Op("Z", "spin"), "X": Op("X", "spin")},
        eps=1e-5,
    )
    te.include_phase = algorithm == "include_phase"

    if algorithm in ["vanilla", "include_phase"]:
        algo = "vqd"
        tau = 0.1
    elif algorithm == "p-VQD":
        algo = "pvqd"
        tau = 0.1
    else:
        algo = "trotter"
        tau = 0.02

    for _ in range(50):
        te.kernel(tau, algo=algo)
    z = te.property_dict["Z"]
    np.testing.assert_allclose(z[:, 0], z[:, 1], atol=1e-2)

    x = te.property_dict["X"]
    np.testing.assert_allclose(x[:, 0], x[:, 1], atol=1e-2)



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from tencirchem.dynamic.model import sbm, pyrazine
from tencirchem.dynamic.transform import qubit_encode_op, qubit_encode_basis
from tencirchem.dynamic.time_derivative import get_ansatz, get_jacobian_func, get_deriv, regularized_inversion
from tencirchem.dynamic.time_evolution import TimeEvolution



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


import logging

import numpy as np
import scipy
import tensorcircuit as tc

from tencirchem.utils.backend import jit
from tencirchem.utils.circuit import evolve_pauli


logger = logging.getLogger(__name__)


def construct_ansatz_op(ham_terms, spin_basis):
    dof_idx_dict = {b.dof: i for i, b in enumerate(spin_basis)}
    ansatz_op_list = []

    for i, op in enumerate(ham_terms):
        logger.info(f"Ansatz operator: {i}, {op}")

        op_mat = 1
        name = ""
        for symbol in op.split_symbol:
            if symbol in ["X", "x", "sigma_x"]:
                op_mat = np.kron(op_mat, tc.gates._x_matrix)
                name += "X"
            elif symbol in ["Y", "y", "sigma_y"]:
                op_mat = np.kron(op_mat, tc.gates._y_matrix)
                name += "Y"
            else:
                if symbol not in ["Z", "z", "sigma_z"]:
                    raise ValueError(f"Hamiltonian must be sum of Pauli strings, got term {op}")
                op_mat = np.kron(op_mat, tc.gates._z_matrix)
                name += "Z"
        qubit_idx_list = [dof_idx_dict[dof] for dof in op.dofs]
        ansatz_op_list.append((op_mat, op.factor, name, qubit_idx_list))

    return ansatz_op_list


def get_circuit(ham_terms, spin_basis, n_layers, init_state, params, param_ids=None, compile_evolution=False):
    if param_ids is None:
        param_ids = list(range(len(ham_terms)))

    params = tc.backend.reshape(params, [n_layers, max(param_ids) + 1])

    ansatz_op_list = construct_ansatz_op(ham_terms, spin_basis)

    if isinstance(init_state, tc.Circuit):
        c = tc.Circuit.from_qir(init_state.to_qir(), circuit_params=init_state.circuit_param)
    else:
        c = tc.Circuit(len(spin_basis), inputs=init_state)

    for i in range(0, n_layers):
        for j, (ansatz_op, _, name, qubit_idx_list) in enumerate(ansatz_op_list):
            param_id = np.abs(param_ids[j])
            # +0.1 is to avoid np.sign(0) problem
            sign = np.sign(param_ids[j] + 0.1)
            theta = sign * params[i, param_id]
            if not compile_evolution:
                np.testing.assert_allclose(ansatz_op @ ansatz_op, np.eye(len(ansatz_op)))
                name = f"exp(-iθ{name})"
                c.exp1(*qubit_idx_list, unitary=ansatz_op, theta=theta, name=name)
            else:
                pauli_string = tuple(zip(qubit_idx_list, name))
                c = evolve_pauli(c, pauli_string, theta=theta)
    return c


def one_trotter_step(ham_terms, spin_basis, init_state, dt, inplace=False):
    """
    one step first order trotter decompostion
    """
    ansatz_op_list = construct_ansatz_op(ham_terms, spin_basis)

    if isinstance(init_state, tc.Circuit):
        if inplace:
            c = init_state
        else:
            c = tc.Circuit.from_qir(init_state.to_qir(), circuit_params=init_state.circuit_param)
    else:
        c = tc.Circuit(len(spin_basis), inputs=init_state)

    for ansatz_op, op_factor, name, qubit_idx_list in ansatz_op_list:
        c.exp1(*qubit_idx_list, unitary=ansatz_op, theta=dt * op_factor, name=name)
    return c


def get_ansatz(ham_terms, spin_basis, n_layers, init_state, param_ids=None):
    @jit
    def ansatz(theta):
        c = get_circuit(ham_terms, spin_basis, n_layers, init_state, theta, param_ids)
        return c.state()

    return ansatz


def get_jacobian_func(ansatz):
    return jit(tc.backend.jacfwd(ansatz, argnums=0))


def regularized_inversion(m, eps):
    evals, evecs = scipy.linalg.eigh(m)
    evals += eps * np.exp(-evals / eps)
    new_evals = 1 / evals
    return evecs @ np.diag(new_evals) @ evecs.T


def regularized_inversion2(m, eps):
    evals, evecs = scipy.linalg.eigh(m)
    mask = np.abs(evals) > 1e-10
    evals = evals[mask]
    assert np.all(evals > 0)
    new_evals = np.zeros(len(evecs))
    new_evals[mask] = 1 / evals
    return evecs @ np.diag(new_evals) @ evecs.T


def get_deriv(ansatz, jacobian_func, params, hamiltonian: np.ndarray, eps: float = 1e-5, include_phase: bool = False):
    params = tc.array_to_tensor(params)

    jacobian = tc.backend.numpy(jacobian_func(params)).astype(np.complex128)
    lhs = jacobian.conj().T @ jacobian

    psi = tc.backend.numpy(ansatz(params)).astype(np.complex128)
    hpsi = hamiltonian @ psi
    rhs = jacobian.conj().T @ hpsi

    lhs = lhs.real
    rhs = rhs.imag

    # global phase term in https://arxiv.org/pdf/1812.08767.pdf
    if include_phase:
        ovlp = jacobian.conj().T @ psi
        lhs += (ovlp.reshape(-1, 1) * ovlp.reshape(1, -1)).real
        e = psi.conj() @ hpsi
        rhs -= (e * ovlp).imag

    lhs_inverse = regularized_inversion(lhs, eps)
    theta_deriv = lhs_inverse @ rhs
    np.testing.assert_allclose(lhs @ lhs_inverse @ rhs, rhs, atol=2e2 * eps)
    np.testing.assert_allclose(theta_deriv.imag, 0)
    return theta_deriv.real.astype(np.float64)


def get_pvqd_loss_func(ansatz):
    def loss(delta_params, params, hamiltonian: np.ndarray, delta_t: float):
        ket = ansatz(params + delta_params)
        bra = ansatz(params)
        evolution = tc.backend.convert_to_tensor(tc.backend.expm(1j * delta_t * hamiltonian))
        return 1 - tc.backend.norm(bra.conj() @ (evolution @ ket)) ** 2

    loss = tc.interfaces.scipy.scipy_optimize_interface(loss)
    return loss



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from renormalizer import BasisHalfSpin, BasisSHO, Op


def get_ham_terms(epsilon, delta, nmode, omega_list, g_list):
    terms = [Op("sigma_z", "spin", epsilon), Op("sigma_x", "spin", delta)]
    for i in range(nmode):
        g = g_list[i]
        omega = omega_list[i]
        terms.extend([Op(r"b^\dagger b", f"v{i}", omega), Op("sigma_z", "spin", g) * Op(r"b^\dagger+b", f"v{i}")])
    return terms


def get_basis(omega_list, nlevels):
    if isinstance(nlevels, int):
        nlevels = [nlevels] * len(omega_list)
    if not isinstance(nlevels, list):
        raise TypeError(f"`nlevels` must be int or list. Got {type(nlevels)}")
    basis = [BasisHalfSpin("spin")]
    for i in range(len(omega_list)):
        basis.append(BasisSHO(f"v{i}", omega=omega_list[i], nbas=nlevels[i]))
    return basis



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


import logging
from itertools import permutations as permut
from itertools import product


import numpy as np
from renormalizer.model import Op
from renormalizer.model.basis import BasisSHO, BasisMultiElectron
from renormalizer.utils.constant import ev2au


logger = logging.getLogger(__name__)


r"""
Bi-linear vibronic coupling model for Pyrazine, 4 modes
See: Raab, Worth, Meyer, Cederbaum.  J.Chem.Phys. 110 (1999) 936
The parameters are from heidelberg mctdh package pyr4+.op
"""


# frequencies
w10a = 0.1139 * ev2au
w6a = 0.0739 * ev2au
w1 = 0.1258 * ev2au
w9a = 0.1525 * ev2au

# energy-gap
delta = 0.42300 * ev2au

# linear, on-diagonal coupling coefficients
# H(1,1)
_6a_s1_s1 = 0.09806 * ev2au
_1_s1_s1 = 0.05033 * ev2au
_9a_s1_s1 = 0.14521 * ev2au
# H(2,2)
_6a_s2_s2 = -0.13545 * ev2au
_1_s2_s2 = 0.17100 * ev2au
_9a_s2_s2 = 0.03746 * ev2au

# quadratic, on-diagonal coupling coefficients
# H(1,1)
_10a_10a_s1_s1 = -0.01159 * ev2au
_6a_6a_s1_s1 = 0.00000 * ev2au
_1_1_s1_s1 = 0.00000 * ev2au
_9a_9a_s1_s1 = 0.00000 * ev2au
# H(2,2)
_10a_10a_s2_s2 = -0.01159 * ev2au
_6a_6a_s2_s2 = 0.00000 * ev2au
_1_1_s2_s2 = 0.00000 * ev2au
_9a_9a_s2_s2 = 0.00000 * ev2au

# bilinear, on-diagonal coupling coefficients
# H(1,1)
_6a_1_s1_s1 = 0.00108 * ev2au
_1_9a_s1_s1 = -0.00474 * ev2au
_6a_9a_s1_s1 = 0.00204 * ev2au
# H(2,2)
_6a_1_s2_s2 = -0.00298 * ev2au
_1_9a_s2_s2 = -0.00155 * ev2au
_6a_9a_s2_s2 = 0.00189 * ev2au

# linear, off-diagonal coupling coefficients
_10a_s1_s2 = 0.20804 * ev2au

# bilinear, off-diagonal coupling coefficients
# H(1,2) and H(2,1)
_1_10a_s1_s2 = 0.00553 * ev2au
_6a_10a_s1_s2 = 0.01000 * ev2au
_9a_10a_s1_s2 = 0.00126 * ev2au

e_list = ["s1", "s2"]
v_list = ["10a", "6a", "9a", "1"]


def get_ham_terms():
    ham_terms = []
    for e_isymbol, e_jsymbol in product(e_list, repeat=2):
        e_op = Op(r"a^\dagger a", [e_isymbol, e_jsymbol])
        for v_isymbol, v_jsymbol in product(v_list, repeat=2):
            # linear
            if v_isymbol == v_jsymbol:
                # if one of the permutation is defined, then the `e_idx_tuple` term should
                # be defined as required by Hermitian Hamiltonian
                for eterm1, eterm2 in permut([e_isymbol, e_jsymbol], 2):
                    factor = globals().get(f"_{v_isymbol}_{eterm1}_{eterm2}")
                    if factor is not None:
                        factor *= np.sqrt(eval(f"w{v_isymbol}"))
                        ham_terms.append(e_op * Op("x", v_isymbol) * factor)
                        logger.debug(f"term: {v_isymbol}_{e_isymbol}_{e_jsymbol}")
                        break
                else:
                    logger.debug(f"no term: {v_isymbol}_{e_isymbol}_{e_jsymbol}")

            # quadratic
            # use product to guarantee `break` breaks the whole loop
            it = product(permut([v_isymbol, v_jsymbol], 2), permut([e_isymbol, e_jsymbol], 2))
            for (vterm1, vterm2), (eterm1, eterm2) in it:
                factor = globals().get(f"_{vterm1}_{vterm2}_{eterm1}_{eterm2}")

                if factor is not None:
                    factor *= np.sqrt(eval(f"w{v_isymbol}") * eval(f"w{v_jsymbol}"))
                    if vterm1 == vterm2:
                        v_op = Op("x^2", vterm1)
                    else:
                        v_op = Op("x", vterm1) * Op("x", vterm2)
                    ham_terms.append(e_op * v_op * factor)
                    logger.debug(f"term: {v_isymbol}_{v_jsymbol}_{e_isymbol}_{e_jsymbol}")
                    break
            else:
                logger.debug(f"no term: {v_isymbol}_{v_jsymbol}_{e_isymbol}_{e_jsymbol}")

    # electronic coupling
    ham_terms.append(Op(r"a^\dagger a", "s1", -delta, [0, 0]))
    ham_terms.append(Op(r"a^\dagger a", "s2", delta, [0, 0]))

    # vibrational kinetic and potential
    for v_isymbol in v_list:
        ham_terms.extend([Op("p^2", v_isymbol, 0.5), Op("x^2", v_isymbol, 0.5 * eval("w" + v_isymbol) ** 2)])

    ham_terms = [term for term in ham_terms if term.factor != 0]

    return ham_terms


def get_basis(nlevels):
    if isinstance(nlevels, int):
        nlevels = [nlevels] * len(v_list)
    if not isinstance(nlevels, list):
        raise TypeError(f"`nlevels` must be int or list. Got {type(nlevels)}")
    basis = [BasisMultiElectron(e_list, [0, 0])]
    for v_isymbol, n in zip(v_list, nlevels):
        basis.append(BasisSHO(v_isymbol, globals()[f"w{v_isymbol}"], n))
    return basis



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


import logging
from functools import partial

import numpy as np
from openfermion import jordan_wigner
import tensorcircuit as tc

from tencirchem.utils.backend import jit
from tencirchem.constants import ad_a_hc2, adad_aa_hc2, ad_a_hc, adad_aa_hc
from tencirchem.utils.misc import ex_op_to_fop
from tencirchem.static.evolve_tensornetwork import get_init_circuit


logger = logging.getLogger(__name__)


@partial(jit, static_argnums=[1, 2, 3, 4, 5])
def get_statevector(params, n_qubits, n_elec, ex_ops, param_ids, hcb=False, init_state=None):
    if tc.backend.name == "jax":
        logger.info(f"Entering `get_statevector`. n_qubit: {n_qubits}")
    if param_ids is None:
        assert len(params) == len(ex_ops)
        param_ids = list(range(len(params)))

    circuit = get_init_circuit(n_qubits, n_elec, hcb, init_state)

    # note only the real part is taken
    statevector = tc.backend.convert_to_tensor(circuit.state())

    for param_id, f_idx in zip(param_ids, ex_ops):
        theta = params[param_id]
        statevector = evolve_excitation(statevector, f_idx, theta, hcb)

    return statevector.real.reshape(-1)


@partial(jit, static_argnums=[1, 3])
def evolve_excitation(statevector, f_idx, theta, hcb):
    n_qubits = round(np.log2(statevector.shape[0]))
    qubit_idx = [n_qubits - 1 - idx for idx in f_idx]
    # fermion operator operated on ket, twice
    f2ket = tc.Circuit(n_qubits, inputs=statevector)
    # fermion operator index, not sorted
    if len(qubit_idx) == 2:
        f2ket.any(*qubit_idx, unitary=ad_a_hc2)
    else:
        assert len(qubit_idx) == 4
        f2ket.any(*qubit_idx, unitary=adad_aa_hc2)

    # fermion operator operated on ket
    fket = apply_excitation(statevector, f_idx, hcb)
    statevector += (1 - tc.backend.cos(theta)) * f2ket.state() + tc.backend.sin(theta) * fket
    return statevector


@partial(jit, static_argnums=[1, 2])
def apply_excitation(statevector, f_idx, hcb):
    n_qubits = round(np.log2(statevector.shape[0]))
    qubit_idx = [n_qubits - 1 - idx for idx in f_idx]
    circuit = tc.Circuit(n_qubits, inputs=statevector)
    # fermion operator index, not sorted
    if len(qubit_idx) == 2:
        circuit.any(*qubit_idx, unitary=ad_a_hc)
    else:
        assert len(qubit_idx) == 4
        circuit.any(*qubit_idx, unitary=adad_aa_hc)

    if hcb:
        return circuit.state()

    # pauli string index, already sorted
    fop = ex_op_to_fop(f_idx)
    qop = jordan_wigner(fop)
    z_indices = []
    for idx, term in next(iter(qop.terms.keys())):
        if term != "Z":
            assert idx in f_idx
            continue
        z_indices.append(n_qubits - 1 - idx)
    if sorted(qop.terms.items())[0][1].real > 0:
        sign = 1
    else:
        sign = -1
    phase_vector = [sign]
    for i in range(n_qubits):
        if i in z_indices:
            phase_vector = np.kron(phase_vector, [1, -1])
        else:
            phase_vector = np.kron(phase_vector, [1, 1])
    return phase_vector * circuit.state()


# external interface
def apply_excitation_statevector(statevector, n_qubits, n_elec, f_idx, hcb):
    return apply_excitation(statevector, f_idx, hcb).real
